"""
"""


#[
from __future__ import annotations
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "solution",
    )

    def __init__(
        self,
        order: int,
        constant: bool,
    ) -> None:
        """
        """
        self.solution = None

    #]





import irispie as ir


source = r"""
!transition_variables
    a, b, c
!unanticipated_shocks
    shk_a, shk_b, shk_c
!transition_equations
    a = shk_a;
    b = shk_b;
    c = shk_c;
"""

m = ir.Simultaneous.from_string(source, linear=True, flatten=True, )

m.assign(
    std_shk_a=0.1,
    std_shk_b=0.2,
    std_shk_c=0.3,
)


simulation_start = ir.qq(2020,1)
simulation_end = ir.qq(2021,4)
simulation_span = simulation_start >> simulation_end


def test_scalar_overwrite():
    db0 = ir.Databox()
    db0["std_shk_a"] = 10
    db0["std_shk_b"] = 20
    db1 = m.vary_stds(None, db0, simulation_span, )
    assert all( i == 10 for i in db1["std_shk_a"].get_values() )
    assert all( i == 20 for i in db1["std_shk_b"].get_values() )
    assert all( i == 0.3 for i in db1["std_shk_c"].get_values() )


def test_scalar_multiplier():
    db0 = ir.Databox()
    db0["std_shk_a"] = 10
    db0["std_shk_b"] = 20
    db1 = m.vary_stds(db0, None, simulation_span, )
    assert all( i == 10*0.1 for i in db1["std_shk_a"].get_values() )
    assert all( i == 20*0.2 for i in db1["std_shk_b"].get_values() )
    assert all( i == 0.3 for i in db1["std_shk_c"].get_values() )


def test_series_overwrite():
    periods = (simulation_start+1, simulation_end-1, )
    db0 = ir.Databox()
    db0["std_shk_a"] = ir.Series(periods=periods, values=[10, 10], )
    db0["std_shk_b"] = ir.Series(periods=periods, values=[20, 20], )
    db1 = m.vary_stds(None, db0, simulation_span, )
    assert db1["std_shk_a"].get_values() ==  (0.1, 10, 0.1, 0.1, 0.1, 0.1, 10, 0.1)
    assert db1["std_shk_b"].get_values() ==  (0.2, 20, 0.2, 0.2, 0.2, 0.2, 20, 0.2)
    assert db1["std_shk_c"].get_values() ==  tuple(0.3 for _ in range(8))


def test_series_multipier():
    periods = (simulation_start+1, simulation_end-1, )
    db0 = ir.Databox()
    db0["std_shk_a"] = ir.Series(periods=periods, values=[10, 10], )
    db0["std_shk_b"] = ir.Series(periods=periods, values=[20, 20], )
    db1 = m.vary_stds(db0, None, simulation_span, )
    assert db1["std_shk_a"].get_values() ==  (0.1, 10*0.1, 0.1, 0.1, 0.1, 0.1, 10*0.1, 0.1)
    assert db1["std_shk_b"].get_values() ==  (0.2, 20*0.2, 0.2, 0.2, 0.2, 0.2, 20*0.2, 0.2)
    assert db1["std_shk_c"].get_values() ==  tuple(0.3 for _ in range(8))


if __name__ == "__main__":
    test_scalar_overwrite()
    test_scalar_multiplier()
    test_series_overwrite()
    test_series_multipier()




"""
Temporal change functions
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
from numbers import (Real, )
import numpy as _np
import documark as _dm

from ..dates import (Span, )
from ._functionalize import FUNC_STRING
from ._categories import CATEGORIES
#]


__all__ = []


class Inlay:
    """
    """
    #[

    @_dm.reference(
        category=None,
        call_name=CATEGORIES["temporal_change"],
        call_name_is_code=False,
        priority=20,
    )
    def temporal_change(
        self,
        by: int | str,
        func: Callable,
        /,
        **kwargs,
    ) -> None:
        r"""
................................................................................

Overview of temporal change functions with flexible time shifts:


| Function      | Description
|---------------|-------------
| `diff`        | First difference, $y_t = x_t - x_s$ 
| `diff_log`    | First difference of logs, $y_t = \log x_t - \log x_s$ 
| `pct`         | Percent change, $y_t = 100 \cdot (x_t/x_s - 1)$ 
| `roc`         | Gross rate of change, $y_t = x_t/x_s$ 


Overview of temporal change functions with fixed time shifts:

| Function      | Description
|---------------|-------------
| `adiff`       | Annualized first difference, $y_t = a \cdot (x_t - x_{t-1})$ 
| `adiff_log`   | Annualized first difference of logs, $y_t = a \cdot (\log x_t - \log x_{t-1})$ 
| `apct`        | Annualized percent change, $y_t = 100 \cdot \left[ (x_t/x_{t-1})^a - 1 \right]$ 
| `aroc`        | Annualized gross rate of change, $y_t = (x_t/x_{t-1})^a$ 

where

* $x_t$ is the value of the time series at $t$;

* $x_s$ is the reference value of the time series in some preceding period $s$;
the value of $s$ depends on the (optional) input argument `shift` (whose default
value is `shift=-1`); see the explanation below;

* $a$ is an annualization factor, determined by the frequency of the time
series (i.e. the number of segments of a give frequency within a year: 1 for
yearly frequency, 2 for semi-annual frequency, 4 for quarterly frequency, 12 for
monthly frequency, 365 for daily frequency).

The `shift` input argument can be either a negative integer or a string to cover
some specific temporal change calculations:

* If `shift` is a negative integer, the reference period is $s := t-k$ where $k$ is
the negative value of `shift`; positive values (leads) of `shift` are not allowed.

* `shift="yoy"` means a year-on-year change, whereby the time shift is set to
the negative of the frequency of the time series, $s := t-a$ with $a$ being
defined above (the annualization factor).

* `shift="soy"` means a change over the start of the current year; in this case,
the reference period $s$ is set to the first segment of the current year (i.e.
the 1st half-year, the 1st quarter, the 1st month, etc.)

* `shift="eopy"` means a change over the end of the previous year; in this case,
the reference period $s$ is set to the last segment of the previous year (i.e.
the 2nd half-year, the 4th quarter, the 12th month, etc.).

* `shift="tty"` means a change throughout the year; this is equivalent to
`shift=-1` except for any start-of-year periods; in start-of-year periods, the
value of the resulting series is unchanged; for instance, `diff` applied to a
quarterly series with `shift="tty"` will return a series in which the Q1 value
will be unchanged, the Q2 value will be the difference between Q2 and Q1, the Q3
value will be the difference between Q3 and Q2, and the Q4 value will be the
difference between Q4 and Q3.


### Functional forms creating a new Series object ###

```
new = irispie.diff(self, shift=-1)
new = irispie.diff_log(self, shift=-1)
new = irispie.pct(self, shift=-1)
new = irispie.roc(self, shift=-1)
new = irispie.adiff(self)
new = irispie.adiff_log(self)
new = irispie.apct(self)
new = irispie.aroc(self)
```


### Class methods changing an existing Series object in-place ###

```
self.diff(shift=-1)
self.difflog(shift=-1)
self.pct(shift=-1)
self.roc(shift=-1)
self.adiff()
self.adifflog()
self.apct()
self.aroc()
```



### Input arguments ###

???+ input "self"
    Time series on whose data a temporal change function is calculated (see
    the overview table above).

???+ input "shift"
    A negative integer or a string determining a time lag at which the temporal change
    function is calculated. If `shift=None` (or not specified), `shift` is
    set to `-1`. If `shift` is a string, it must be one of the following:

    * `"yoy"`: year-on-year change
    * `"soy"`: change over the start of the current year
    * `"eopy"`: change over the end of the previous year
    * `"tty"`: change throughout the year


### Returns ###

???+ returns "new"
    New time series with data calculated as a temporal change function of
    the original data.

???+ returns "self"
    Time series with data replaced by a temporal change function of the
    original data.

................................................................................
        """
        _catch_invalid_shift(by, )
        other = self.copy()
        other.shift(by, **kwargs, )
        self._binop(other, func, new=self, )

    @_dm.reference(
        category=None,
        call_name=CATEGORIES["temporal_change_conversion"],
        call_name_is_code=False,
        priority=19,
    )
    def temporal_change_conversion(self, ):
        r"""
................................................................................

Overview of temporal change conversions:

| Function          | Description
|-------------------|-------------
| `roc_from_pct`    | Convert percent change to gross rate of change
| `pct_from_roc`    | Convert gross rate of change to percent change
| `pct_from_apct`   | Convert annualized percent change to percent change
| `roc_from_apct`   | Convert annualized percent change to gross rate of change
| `roc_from_aroc`   | Convert annualized gross rate of change to gross rate of change

### Functional forms changing an existing Series object in-place ###

```
new = irispie.roc_from_pct(self)
new = irispie.pct_from_roc(self)
new = irispie.pct_from_apct(self)
new = irispie.roc_from_apct(self)
new = irispie.roc_from_aroc(self)
```

### Class methods changing an existing Series object in-place ###

```
self.roc_from_pct()
self.pct_from_roc()
self.pct_from_apct()
self.roc_from_apct()
self.roc_from_aroc()
```

................................................................................
        """
        pass

    @_dm.reference(category="temporal_change", )
    def diff(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
................................................................................

==First difference==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        self.temporal_change(shift, lambda x, y: x - y, neutral_value=0, )

    @_dm.reference(category="temporal_change", )
    def adiff(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Annualized first difference==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: factor*(x - y), neutral_value=0, )

    @_dm.reference(category="temporal_change", )
    def diff_log(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
................................................................................

==First difference of logs==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        self.temporal_change(shift, lambda x, y: _np.log(x) - _np.log(y), neutral_value=0, )

    @_dm.reference(category="temporal_change", )
    def adiff_log(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Annualized first difference of logs==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: factor*(_np.log(x) - _np.log(y)), neutral_value=0, )

    @_dm.reference(category="temporal_change", )
    def roc(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
................................................................................

==Gross rate of change==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        self.temporal_change(shift, lambda x, y: x/y, neutral_value=1, )

    @_dm.reference(category="temporal_change", )
    def aroc(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Annualized gross rate of change==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: (x/y)**factor, neutral_value=1, )

    @_dm.reference(category="temporal_change", )
    def pct(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
................................................................................

==Percent change==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        self.temporal_change(shift, lambda x, y: 100*(x/y - 1), neutral_value=None, )

    @_dm.reference(category="temporal_change", )
    def apct(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Annualized percent change==

See documentation for [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

................................................................................
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: 100*((x/y)**factor - 1), neutral_value=None, )

    @_dm.reference(category="temporal_change_conversion", )
    def roc_from_pct(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Gross rate of change from percent change==

See documentation for [converting measures of temporal
change](#temporal-change-conversion).

................................................................................
        """
        self.data = 1 + self.data/100

    @_dm.reference(category="temporal_change_conversion", )
    def pct_from_roc(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Percent change from gross rate of change==

See documentation for [converting measures of temporal
change](#temporal-change-conversion).

................................................................................
        """
        self.data = 100*(self.data - 1)

    @_dm.reference(category="temporal_change_conversion", )
    def pct_from_apct(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Percent change from annualized percent change==

See documentation for [converting measures of temporal
change](#temporal-change-conversion).

................................................................................
        """
        factor = self.frequency.value or 1
        self.data = 100*((1 + self.data/100)**(1/factor) - 1)

    @_dm.reference(category="temporal_change_conversion", )
    def roc_from_apct(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Gross rate of change from annualized percent change==

See documentation for [converting measures of temporal
change](#temporal-change-conversion).

................................................................................
        """
        factor = self.frequency.value or 1
        self.data = (1 + self.data/100)**(1/factor)

    @_dm.reference(category="temporal_change_conversion", )
    def roc_from_aroc(
        self,
        /,
    ) -> None:
        r"""
................................................................................

==Gross rate of change from annualized gross rate of change==

See documentation for [converting measures of temporal
change](#temporal-change-conversion).

................................................................................
        """
        factor = self.frequency.value or 1
        self.data = self.data**factor

    @_dm.reference(category="temporal_cumulation", )
    def cum_diff(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Cumulation of first differences==

See documentation for [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

................................................................................
        """
        self.temporal_cumulation("diff", *args, **kwargs, )

    @_dm.reference(category="temporal_cumulation", )
    def cum_diff_log(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Cumulation of first differences of logs==

See documentation for [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

................................................................................
        """
        self.temporal_cumulation("diff_log", *args, **kwargs, )

    @_dm.reference(category="temporal_cumulation", )
    def cum_pct(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Cumulation of percent changes==

See documentation for [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

................................................................................
        """
        self.temporal_cumulation("pct", *args, **kwargs, )

    @_dm.reference(category="temporal_cumulation", )
    def cum_roc(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Cumulation of gross rates of change==

See documentation for [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

................................................................................
        """
        self.temporal_cumulation("roc", *args, **kwargs, )

    @_dm.reference(
        category=None,
        call_name=CATEGORIES["temporal_cumulation"],
        call_name_is_code=False,
        priority=20,
    )
    def temporal_cumulation(
        self,
        func_name: str,
        shift: int | str = -1,
        initial: Real | Self | None = None,
        span: Span | None = None,
    ) -> None:
        r"""
................................................................................

Overview of temporal cumulation calculations:

| Function      | Description
|---------------|-------------
| `cum_diff`    | Cumulative difference, $y_t = \sum_{i=0}^{t} (x_i - x_{i-k})$
| `cum_diff_log`| Cumulative difference of logs, $y_t = \sum_{i=0}^{t} (\log x_i - \log x_{i-k})$
| `cum_pct`     | Cumulative percent change, $y_t = 100 \sum_{i=0}^{t} (x_i/x_{i-k} - 1)$
| `cum_roc`     | Cumulative gross rate of change, $y_t = \prod_{i=0}^{t} (x_i/x_{i-k})$

where

* $k$ is a time shift (time lag) with which the temporal change is
calculated, determined by the negative value of the `shift` input argument.

* $t$ is the end date of the time series.


### Functional forms creating new Series objects ###

```
new = irispie.cum_diff(self, shift=-1, initial=None, span=None)
new = irispie.cum_diff_log(self, shift=-1, initial=None, span=None)
new = irispie.cum_pct(self, shift=-1, initial=None, span=None)
new = irispie.cum_roc(self, shift=-1, initial=None, span=None)
```


### Class methods changing existing Series objects in-place ###

```
self.cum_diff(shift=-1, initial=None, span=None)
self.cum_diff_log(shift=-1, initial=None, span=None)
self.cum_pct(shift=-1, initial=None, span=None)
self.cum_roc(shift=-1, initial=None, span=None)
```

### Input arguments ###

???+ input "self"
    Time series on whose data a temporal cumulation is calculated (see the
    overview table above).

???+ input "shift"
    A negative integer determining a time lag at which the temporal
    cumulation is calculated. If `shift=None` (or not specified), `shift`
    is set to `-1`.

???+ input "initial"
    Initial value of the cumulative series. If `initial=None` (or not specified),
    the initial value is set to `0` for `diff` and `diff_log`, and to
    `1` for `pct` and `roc`.

???+ input "span"
    Time span on which the values from the original series are cumulated.
    If `span=None` (or not specified), the time span is set to the entire
    time series.


### Returns ###

???+ returns "new"
    New time series with data calculated as temporal cumulation of the
    original data.

???+ returns "self"
    Time series with data replaced by temporal cumulation of the original
    data.

................................................................................
        """
        _catch_invalid_shift(shift, )
        span = Span(None, None, ) if span is None else span
        span = span.resolve(self, )
        direction = span.direction
        factory = _CUMULATIVE_FACTORY[func_name]
        cum_func = factory[direction]
        initial = factory["initial"] if initial is None else initial
        if direction == "forward":
            self._cumulate_forward(shift, cum_func, initial, span, )
        elif direction == "backward":
            self._cumulate_backward(shift, cum_func, initial, span, )

    def _cumulate_forward(self, shift, cum_func, initial, span, /, ) -> None:
        """
        """
        zipped_span = tuple((t, t.shift(shift, )) for t in span)
        zipped_span = tuple((t, sh) for t, sh in zipped_span if sh is not None)
        min_period = min((sh for t, sh in zipped_span), default=span.start_date, )
        initial_span = Span(min_period, span.end_date, )
        change = self.copy()
        self.empty()
        self.set_data(initial_span, initial)
        for t, sh in zipped_span:
            new_data = cum_func(self.get_data(sh, ), change.get_data(t, ), )
            self.set_data(t, new_data)

    def _cumulate_backward(self, shift, cum_func, initial, shifted_backward_range, /, ) -> None:
        """
        """
        orig_range_shifted = Span(self.start_date, self.end_date, -1, )
        orig_range_shifted.shift(shift, )
        shifted_backward_range = shifted_backward_range.resolve(orig_range_shifted, )
        backward_range = shifted_backward_range.copy()
        backward_range.shift(-shift, )
        initial_span = Span(min(shifted_backward_range), backward_range.start_date, )
        orig = self.copy()
        self.empty()
        self.set_data(initial_span, initial, )
        for t, sh in zip(backward_range, shifted_backward_range, ):
            new_data = cum_func(self.get_data(t, ), orig.get_data(t, ), )
            self.set_data(sh, new_data, )

    #]


attributes = (n for n in dir(Inlay) if not n.startswith("_"))
for n in attributes:
    exec(FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__.append(n)


_CUMULATIVE_FACTORY = {
    "diff": {
        "forward": lambda x_past, change_curr: x_past + change_curr,
        "backward": lambda x_future, change_future: x_future - change_future,
        "initial": 0,
    },
    "diff_log": {
        "forward": lambda x_past, change_curr: x_past * exp(change_curr),
        "backward": lambda x_future, change_future: x_future / exp(change_future),
        "initial": 0,
    },
    "pct": {
        "forward": lambda x_past, change_curr: x_past * (1 + change_curr/100),
        "backward": lambda x_future, change_future: x_future / (1 + change_future/100),
        "initial": 1,
    },
    "roc": {
        "forward": lambda x_past, change_curr: x_past * change_curr,
        "backward": lambda x_future, change_future: x_future / change_future,
        "initial": 1,
    },
}


def _catch_invalid_shift(shift: int | str, ):
    """
    """
    if not isinstance(shift, str) and (int(shift) != shift or shift >= 0):
        raise ValueError("Time shift must be a negative integer or a string")


def _roc_from_pct(pct: Real, /, ) -> Real:
    """
    """
    return 1 + pct/100





"""
Implement SteadyDataboxableProtocol
"""


#[

from __future__ import annotations

import numpy as _np

from .. import has_variants as _has_variants
from ..quantities import QuantityKind
from ..series.main import Series

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from numbers import Real
    from ..dates import Period

#]


"""
Quantities that are time series in model databoxes
"""
_TIME_SERIES_QUANTITY = QuantityKind.ANY_VARIABLE | QuantityKind.ANY_SHOCK


# REFACTOR:
# Create a SteadyBoxable class
class Inlay:
    def generate_steady_items(
        self,
        start: Period,
        end: Period,
        /,
        deviation: bool = False,
        unpack_singleton: bool = True,
    ) -> Iterable[tuple[str, Series | Real]]:
        """
        """
        num_periods = int(end - start + 1)
        shift_in_first_column = start.get_distance_from_origin()
        qid_to_name = self.create_qid_to_name()
        qid_to_description = self.create_qid_to_description()
        #
        tuple_of_arrays = tuple(
            self.create_some_array(
            variant=variant,
            deviation=deviation,
            num_columns=num_periods,
            shift_in_first_column=shift_in_first_column,
        ) for variant in self._variants)
        array = _np.stack(arrays=tuple_of_arrays, axis=2, )
        #
        num_rows = array.shape[0]
        qid_to_kind = self.create_qid_to_kind()
        for qid in qid_to_name.keys():
            name = qid_to_name[qid]
            array_slice = array[qid, :, :]
            is_time_series = qid_to_kind[qid] in _TIME_SERIES_QUANTITY
            if is_time_series:
                value = Series(
                    start=start,
                    values=array_slice,
                    description=qid_to_description[qid],
                )
            else:
                value = _has_variants.unpack_singleton(
                    array_slice[0, :].tolist(), self.is_singleton,
                    unpack_singleton=unpack_singleton,
                )
            yield name, value,




"""
Cells in simulation dataslates
"""


#[
from __future__ import annotations

from typing import (NamedTuple, Self, )
#]


class Cell(NamedTuple, ):
    """
    Cell in a simulation dataslate
    """
    #[

    qid: int
    column: int

    def shifted(
        self: Self,
        by: int,
    ) -> Self:
        """
        Create a new Cell with a shifted column
        """
        return Cell(self.qid, self.column+by, )

    #]





"""
Initialize median and MSE matrix for alpha vector
"""


#[
from __future__ import annotations

import numpy as _np
from numbers import (Real, )

from .solutions import (Solution, left_div, )
from . import covariances as _covariances
#]


_DEFAULT_DIFFUSE_SCALE = 1e8


def _approx_diffuse(
        solution: Solution,
        custom_diffuse_scale: Real | None = None,
) -> tuple[Real, _np.ndarray | None]:
    """
    """
    #[
    diffuse_scale = custom_diffuse_scale or _DEFAULT_DIFFUSE_SCALE
    unknown_init_impact = None
    return diffuse_scale, unknown_init_impact
    #]


def _fixed_unknown(
        solution: Solution,
        custom_diffuse_scale: Real | None = None,
) -> tuple[Real, _np.ndarray | None]:
    """
    """
    #[
    diffuse_scale = 0
    unknown_init_impact = (
        _np.eye(solution.num_xi, solution.num_unit_roots, )
        if solution.num_unit_roots else None
    )
    return diffuse_scale, unknown_init_impact,
    #]


def _fixed_zero(
        solution: Solution,
        custom_diffuse_scale: Real | None = None,
) -> tuple[Real, _np.ndarray | None]:
    """
    """
    #[
    diffuse_scale = 0
    unknown_init_impact = None
    return diffuse_scale, unknown_init_impact,
    #]


_RESOLVE_DIFFUSE = {
    "approx_diffuse": _approx_diffuse,
    "fixed_unknown": _fixed_unknown,
    "fixed_zero": _fixed_zero,
}


def initialize(
    solution: Solution,
    cov_u: _np.ndarray,
    *,
    diffuse_method: Literal["approx_diffuse", "fixed_unknown", "fixed_zero", ] = "fixed_unknown",
    diffuse_scale: Real | None = None,
) -> tuple[np_.ndarray, np_.ndarray, _np.ndarray, ]:
    r"""
    Return median and MSE matrix for initial alpha, and the impact of fixed
    unknowns on initial alpha
    """
    #[
    diffuse_func = _RESOLVE_DIFFUSE[diffuse_method]
    diffuse_scale, unknown_init_impact = diffuse_func(solution, diffuse_scale, )
    init_med = _initialize_med(solution, )
    init_mse = _initialize_mse(solution, cov_u, diffuse_scale, )
    return init_med, init_mse, unknown_init_impact,
    #]


def _initialize_med(solution: Solution, ) -> _np.ndarray:
    """
    Solve alpha_stable = Ta_stable @ alpha_stable + Ka_stable for alpha_stable
    and return alpha with 0s for unstable elements
    """
    #[
    num_alpha = solution.num_alpha
    num_unit_roots = solution.num_unit_roots
    num_stable = solution.num_stable
    Ta_stable = solution.Ta_stable
    Ka_stable = solution.Ka_stable
    init_med = _np.zeros((num_alpha, ), dtype=float, )
    #
    T = _np.eye(num_stable, dtype=float, ) - Ta_stable
    init_med_stable = left_div(T, Ka_stable, )
    init_med[num_unit_roots:] = init_med_stable
    return init_med
    #]


def _initialize_mse(
    solution: Solution,
    cov_u: _np.ndarray,
    diffuse_scale: Real | None = None,
) -> _np.ndarray:
    """
    """
    #[
    init_mse = _covariances.get_cov_alpha_00(solution, cov_u, )
    if diffuse_scale:
        num_unit_roots = solution.num_unit_roots
        init_mse[:num_unit_roots, :num_unit_roots] = \
            _initialize_mse_unstable_approx_diffuse(solution, cov_u, init_mse, diffuse_scale, )
    init_mse = _covariances.symmetrize(init_mse, )
    return init_mse
    #]


def _initialize_mse_unstable_approx_diffuse(
    solution: Solution,
    cov_u: _np.ndarray,
    init_mse: _np.ndarray,
    diffuse_scale: Real | None = None,
) -> _np.ndarray:
    """
    """
    #[
    num_unit_roots = solution.num_unit_roots
    num_alpha = solution.num_alpha
    base_cov = (
        init_mse[num_unit_roots:, num_unit_roots:] if num_alpha
        else (cov_u if cov_u.size else _np.ones((1, 1, ), dtype=float, ))
    )
    scale = _mean_of_diag(base_cov, ) * diffuse_scale
    return scale * _np.eye(num_unit_roots, dtype=float, )
    #]


def _mean_of_diag(x: _np.ndarray, ) -> Real:
    """
    """
    return _np.mean(_np.diag(x, ), )




"""
Implement SlatableProtocol
"""


#[

from __future__ import annotations

import warnings as _wa

from .. import slatables as _slatables
from ..series.main import Series
from .. import quantities as _quantities

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from .main import Stacker

#]


_DEFAULT_SHOCK_VALUE = 0.0


class Slatable(_slatables.Slatable):
    r"""
    """
    #[

    def __init__(
        self,
        stacker: Stacker,
        **kwargs,
    ) -> Self:
        r"""
        """
        #
        if stacker.num_variants != 1:
            raise ValueError("Not implemented for multiple variants")
        #
        super().__init__(**kwargs, )
        self.max_lag = 0 # stacker.max_lag
        self.max_lead = 0 # stacker.max_lead
        #
        qid_to_name = stacker.create_qid_to_name()
        self.databox_names = tuple(
            qid_to_name[qid]
            for qid in sorted(qid_to_name)
        )
        #
        name_to_description = stacker.create_name_to_description()
        self.descriptions = tuple(
            name_to_description.get(name, "", )
            for name in self.databox_names
        )
        #
        # Databox validation - all variables must be time series
        variable_names = stacker.get_names(kind=_quantities.ANY_VARIABLE, )
        validator = (
            lambda x: isinstance(x, Series),
            "Input data for this variable is not a time series",
        )
        self.databox_validators = {
            name: validator
            for name in variable_names
        }
        #
        # Fallbacks and overwrites
        self.fallbacks = {}
        self.overwrites = {}
        #
        shock_names = stacker.get_names(kind=_quantities.ANY_SHOCK, )
        shock_name_to_value = {
            name: [_DEFAULT_SHOCK_VALUE, ]*stacker.num_variants
            for name in shock_names
        }
        if self.shocks_from_data:
            self.fallbacks.update(shock_name_to_value, )
        else:
            self.overwrites.update(shock_name_to_value, )
        #
        stds = stacker._variants[0].std_name_to_value
        if self.stds_from_data:
            self.fallbacks.update(stds, )
        else:
            self.overwrites.update(stds, )
        #
        self.qid_to_logly = stacker.create_qid_to_logly()

    #]




"""
"""


#[

from __future__ import annotations

import warnings as _wa
import copy as _co
import numpy as _np
import operator as _op

from ..conveniences import copies as _copies
from .. import quantities as _quantities

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real
    from typing import Self, Literal, Callable
    from collections.abc import Iterable

#]


class Variant:
    """
    Container for parameter variant specific attributes of a model
    """
    #[

    __slots__ = (
        "levels",
        "changes",
        "solution",
    )

    def __init__(self, **kwargs, ) -> None:
        """
        """
        for n in self.__slots__:
            setattr(self, n, None, )

    @classmethod
    def from_source(
        klass,
        quantities: Iterable[_quantities.Quantity],
        is_flat: bool,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        max_qid = _quantities.get_max_qid(quantities, )
        qid_range = range(max_qid+1, )
        self._initilize_values(qid_range, )
        if is_flat:
            qid_to_logly = _quantities.create_qid_to_logly(quantities, )
            self.zero_changes(qid_to_logly, )
        return self

    @property
    def _all_qids(self, ) -> Iterable[int]:
        """
        """
        return self.levels.keys()

    def copy(self, ) -> Self:
        """
        """
        new = type(self)()
        for i in ("levels", "changes", "solution", ):
            attr = getattr(self, i, )
            if attr is not None:
                setattr(new, i, attr.copy(), )
        return new

    def _initilize_values(self, qid_range, ) -> None:
        """
        """
        self.levels = { qid: None for qid in qid_range }
        self.changes = { qid: None for qid in qid_range }

    def update_values_from_dict(self, update: dict, ) -> None:
        _update_from_dict(self.levels, update, _op.itemgetter(0), lambda x: x, )
        _update_from_dict(self.changes, update, _op.itemgetter(1), lambda x: ..., )

    def update_levels_from_array(self, levels: _np.ndarray, qids: Iterable[int], ) -> None:
        _update_from_array(self.levels, levels, qids, )

    def update_changes_from_array(self, changes: _np.ndarray, qids: Iterable[int], ) -> None:
        _update_from_array(self.changes, changes, qids, )

    def retrieve_values_as_array(
        self,
        attr: Literal["levels", "changes"],
        qids: Iterable[int] | None = None,
    ) -> _np.ndarray:
        values = getattr(self, attr, )
        qids = qids if qids is not None else values.keys()
        return _np.array(tuple(values.get(qid, None, ) for qid in qids), dtype=float, )

    def rescale_values(
        self,
        attr: Literal["levels", "changes"],
        factor: Real,
        qids: Iterable[int] | None = None,
    ) -> None:
        """
        Rescale values by a common factor
        """
        values = getattr(self, attr, )
        qids = qids if qids is not None else values.keys()
        for qid in qids:
            values[qid] *= factor if values[qid] is not None else None

    def retrieve_levels_as_dict(
        self,
        qids: Iterable[int],
    ) -> dict[int, Real | None]:
        r"""
        """
        return { i: self.levels[i] for i in qids }

    def retrieve_changes_as_dict(
        self,
        qids: Iterable[int],
    ) -> dict[int, Real | None]:
        r"""
        """
        return { i: self.changes[i] for i in qids }

    def retrieve_maybelog_values_for_qids(
        self,
        qids: Iterable[int],
        qid_to_logly: dict[int, bool],
    ) -> tuple[_np.ndarray, _np.ndarray, ]:
        """
        """
        #
        # Extract levels and changes as arrays
        qids = tuple(qids)
        maybelog_levels = self.retrieve_values_as_array("levels", qids, )
        maybelog_changes = self.retrieve_values_as_array("changes", qids, )
        #
        # Logarithmize
        where_logly = list(_quantities.generate_where_logly(qids, qid_to_logly, ))
        maybelog_levels[where_logly] = _np.log(maybelog_levels[where_logly])
        maybelog_changes[where_logly] = _np.log(maybelog_changes[where_logly])
        #
        return maybelog_levels, maybelog_changes,

    def zero_changes(
        self,
        qid_to_logly: dict[int, bool | None],
    ) -> None:
        """
        Reset all quantities to flat
        """
        for qid in self.changes.keys():
            logly = qid_to_logly.get(qid, None, )
            self.changes[qid] = float(logly) if logly is not None else None

    def create_steady_array(
        self,
        qid_to_logly: dict[int, bool | None],
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = self.retrieve_values_as_array("levels", ).reshape(-1, 1)
        if num_columns==1 and shift_in_first_column==0:
            return levels
        changes = self.retrieve_values_as_array("changes", ).reshape(-1, 1)
        #
        where_logly = list(_quantities.generate_where_logly(self._all_qids, qid_to_logly))
        #
        shift_vec = _np.array(range(shift_in_first_column, shift_in_first_column+num_columns))
        #
        _wa.filterwarnings(action="ignore", category=RuntimeWarning)
        levels[where_logly] = _np.log(levels[where_logly])
        changes[where_logly] = _np.log(changes[where_logly])
        _wa.filterwarnings(action="default", category=RuntimeWarning)
        #
        levels[_np.isnan(levels) | _np.isinf(levels)] = _np.nan
        changes[_np.isnan(changes) | _np.isinf(changes)] = 0
        #
        steady_array = levels + changes * shift_vec
        #
        _wa.filterwarnings(action="ignore", category=RuntimeWarning)
        steady_array[where_logly, :] = _np.exp(steady_array[where_logly, :])
        _wa.filterwarnings(action="default", category=RuntimeWarning)
        #
        return steady_array

    def create_zero_array(
        self,
        qid_to_logly: dict[int, bool],
        #
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = self.retrieve_values_as_array("levels", ).reshape(-1, 1)
        inx_set_to_0 = [ qid for qid, logly in qid_to_logly.items() if logly is False ]
        inx_set_to_1 = [ qid for qid, logly in qid_to_logly.items() if logly is True ]
        levels[inx_set_to_0] = 0
        levels[inx_set_to_1] = 1
        return _np.tile(levels, (1, num_columns, ))

    def to_portable(self, qid_to_name, ) -> dict[str, Any]:
        """
        """
        return {
            qid_to_name[qid]: (level, self.changes[qid], )
            for qid, level in self.levels.items()
        }


def _update_from_array(
    what_to_update: dict[int, Real | None],
    updated_values: _np.ndarray | Iterable[Real] | None,
    qids: Iterable[int],
) -> None:
    """
    Update levels or changes from an array and a list of qids
    """
    #[
    if updated_values is None:
        return
    if hasattr(updated_values, "flatten"):
        updated_values = updated_values.flatten().tolist()
    for qid, value in zip(qids, updated_values):
        what_to_update[qid] = value if not _is_nan(value) else None
    #]


def _is_nan(x: Real | None, ) -> bool:
    return x != x


def _update_from_dict(
    what_to_update: dict[int, Real | None],
    update: dict[int, Real | tuple[Real]],
    when_tuple: Callable,
    when_not_tuple: Callable,
) -> _np.ndarray:
    """
    Update levels or changes from a dictionary
    """
    #[
    for qid, value in update.items():
        new_value = when_tuple(value) if isinstance(value, tuple) else when_not_tuple(value)
        new_value = new_value if new_value is not ... else what_to_update[qid]
        what_to_update[qid] = new_value if not _is_nan(new_value) else None
    #]






import irispie as ir

db = ir.Databox()
db["x"] = ir.Series(start=ir.qq(2020,1), values=range(20))

fig = ir.make_subplots((2,3), )

db["x"].plot(figure=fig, subplot=0, show_figure=False, )
db["x"].plot(figure=fig, subplot=1, show_figure=False, )





import importlib.metadata as _md

from .dates import *
from .dates import __all__ as dates_all

from .series import *
from .series import __all__ as series_all

from .ez_plotly import *
from .ez_plotly import __all__ as ez_plotly_all

from .rephrases import *
from .rephrases import __all__ as rephrases_all

from .databoxes.main import *
from .databoxes.main import __all__ as databoxes_all

from .chartpacks.main import *
from .chartpacks.main import __all__ as chartpacks_all

from .dataslates.main import *
from .dataslates.main import __all__ as dataslates_all

from .sources import *
from .sources import __all__ as sources_all

from .simultaneous import *
from .simultaneous import __all__ as simultaneous_all

from .stackers import *
from .stackers import __all__ as stackers_all

from .fords import *
from .fords import __all__ as fords_all

from .vars import *
from .vars import __all__ as vars_all

from .quantities import *
from .quantities import __all__ as quantities_all

from .equations import *
from .equations import __all__ as equations_all

from .sequentials import *
from .sequentials import __all__ as sequentials_all

from .explanatories import *
from .explanatories import __all__ as explanatories_all

from .plans import *
from .plans import __all__ as plans_all

from .namings import *
from .namings import __all__ as namings_all

from .file_io import *
from .file_io import __all__ as file_io_all

from .portables import *
from .portables import __all__ as portables_all


__version__ = _md.version(__name__)
__doc__ = _md.metadata(__name__).json["description"]


def print_readme():
    print(__doc__)


#[
def min_irispie_version_required(
    minimum_version_str: str,
):
    """
    """
    current_version = _convert_version(__version__, )
    minimum_version = _convert_version(minimum_version_str, )
    if current_version < minimum_version:
        raise Exception(
            f"Current version of irispie ({__version__}) is less than minimum version required ({minimum_version_str})"
        )


def _convert_version(version_str: str) -> tuple[int, ...]:
    return tuple(int(s) for s in version_str.split("."))
#]


__all__ = (
    *dates_all,
    *series_all,
    *ez_plotly_all,
    *rephrases_all,
    *databoxes_all,
    *sources_all,
    *simultaneous_all,
    *stackers_all,
    *fords_all,
    *vars_all,
    *quantities_all,
    *equations_all,
    *sequentials_all,
    *explanatories_all,
    *plans_all,
    *namings_all,
    *file_io_all,
    *portables_all,
    "min_irispie_version_required",
    "__version__",
)




"""
Algorithmic differentiator
"""


#[
from __future__ import annotations

from typing import (Self, Callable, Protocol, )
from numbers import Real
from collections.abc import (Iterable, Sequence, )
import numpy as _np
import scipy as _sp
import copy as _cp

from ..incidences.main import (Token, )
from ..equators import plain as _equators
from .. import equations as _equations
from . import finite_differentiators as af_
from . import adaptations as _adaptations
#]


class ValueMixin:
    """
    """
    #[

    #]


class Atom(ValueMixin, ):
    """
    Atomic value for differentiation
    """
    #[

    _data_context: _np.ndarray | None = None
    _column_offset: int = 0
    _is_atom: bool = True

    __slots__ = (
        '_value',
        '_diff',
        '_logly',
        '_data_index',
        '_row_index',
        '_column_index',
    )

    def __init__(self) -> None:
        """
        """
        self._value = None
        self._diff = None
        self._logly = None
        self._data_index = None
        self._row_index = None
        self._column_index = None

    @classmethod
    def no_context(
        klass,
        value: Real,
        diff: Real,
        /,
        logly: bool | None = False,
    ) -> Self:
        """
        Create atom with self-contained data and logly contexts
        """
        self = klass()
        self._value = value
        self._diff = diff
        self._logly = logly if logly is not None else False
        return self

    @classmethod
    def in_context(
        klass,
        /,
        diff: _np.ndarray | Real,
        data_index: tuple[int, slice],
        logly: bool,
    ) -> Self:
        """
        Create atom with pointers to data and logly contexts
        """
        self = Atom()
        self._diff = diff
        self._data_index = data_index
        self._row_index = data_index[0]
        self._column_index = data_index[1]
        self._logly = logly if logly is not None else False
        return self

    @classmethod
    def zero(
        klass,
        diff_shape: tuple[int, int],
        /,
    ) -> Self:
        return klass.no_context(0, _np.zeros(diff_shape, dtype=_np.float64), False)

    @property
    def value(self, /, ):
        if self._value is not None:
            return self._value
        else:
            column_index = self._column_index + self._column_offset
            return self._data_context[self._row_index, column_index]

    @property
    def diff(self):
        return self._diff if not self._logly else self._diff * self.value

    def __pos__(self):
        return self

    def __neg__(self):
        new_value = -self.value
        new_diff = -self.diff
        return type(self).no_context(new_value, new_diff, False)

    def __add__(self, other):
        if hasattr(other, "_is_atom"):
            """
            Differentiate self(x)+other(x)
            """
            new_value = self.value + other.value
            new_diff = self.diff + other.diff
        else:
            """
            Differentiate self(x)+other
            """
            new_value = self.value + other
            new_diff = self.diff
        return type(self).no_context(new_value, new_diff, False)

    def __sub__(self, other):
        if hasattr(other, "_is_atom"):
            """
            Differentiate self(x)-other(x)
            """
            new_value = self.value - other.value
            new_diff = self.diff - other.diff
        else:
            """
            Differentiate self(x)-other
            """
            new_value = self.value - other
            new_diff = self.diff
        return type(self).no_context(new_value, new_diff, False)

    def __mul__(self, other):
        self_value = self.value
        self_diff = self.diff
        if hasattr(other, "_is_atom"):
            """
            Differentiate self(x)*other(x)
            """
            other_value = other.value
            other_diff = other.diff
            new_value = self_value * other_value
            new_diff = self_diff * other_value + self_value * other_diff
        else:
            """
            Differentiate self(x)*other
            """
            new_value = self_value * other
            new_diff = self_diff * other
        return type(self).no_context(new_value, new_diff, False)

    def __truediv__(self, other):
        self_value = self.value
        self_diff = self.diff
        if hasattr(other, "_is_atom"):
            other_value = other.value
            other_diff = other.diff
            new_value = self_value / other_value
            new_diff = (self_diff*other_value - self_value*other_diff) / (other_value**2)
        else:
            new_value = self_value / other
            new_diff = self_diff / other
        return type(self).no_context(new_value, new_diff, False)

    def __rtruediv__(self, other):
        self_value = self.value
        self_diff = self.diff
        new_value = other / self_value
        new_diff = -other*self_diff / (self_value**2)
        return type(self).no_context(new_value, new_diff, False)

    def __pow__(self, other):
        """
        Differenatiate self**other 
        """
        if hasattr(other, "_is_atom"):
            # self(x)**other(x)
            # d[self(x)**other(x)] = d[self(x)**other] + d[self**other(x)]
            new_value = self.value**other.value
            _, new_diff_exp = self._power(other.value) # self(x)**other
            _, new_diff_pow = other._exponential(self.value) # self**other(x)
            new_diff = new_diff_exp + new_diff_pow
        else:
            # self(x)**other
            new_value, new_diff = self._power(other)
        return type(self).no_context(new_value, new_diff, False)

    def _exponential(self, other_value):
        """
        Differenatiate exponential function other**self(x)
        """
        new_value = other_value**self.value
        new_diff = other_value**self.value * _np.log(other_value) * self.diff 
        return new_value, new_diff

    def _power(self, other_value):
        """
        Differenatiate power function self(x)**other
        """
        self_value = self.value
        self_diff = self.diff
        new_value = self_value ** other_value
        new_diff = other_value * (self_value**(other_value-1)) * self_diff
        return new_value, new_diff

    __rmul__ = __mul__

    __radd__ = __add__

    def __rsub__(self, other, /, ) -> Self:
        """
        Differenatiate other - self
        """
        return self.__neg__().__add__(other)

    def log(self, /, ) -> Self:
        """
        Differentiate log(self)
        """
        new_value = _np.log(self.value)
        new_diff = 1 / self.value * self.diff
        return type(self).no_context(new_value, new_diff, False)

    def exp(self, /, ) -> Self:
        """
        Differentiate exp(self)
        """
        new_value = _np.exp(self.value)
        new_diff = new_value * self.diff
        return type(self).no_context(new_value, new_diff, False)

    def sqrt(self, /, ) -> Self:
        """
        Differentiate sqrt(self)
        """
        new_value = _np.sqrt(self.value)
        new_diff = 0.5 / _np.sqrt(self.diff)
        return type(self).no_context(new_value, new_diff, False)

    def logistic(self, /, ) -> Self:
        """
        Differentiate logistic(self)
        """
        new_value = _sp.special.expit(self.value, )
        new_diff = new_value * (1 - new_value) * self.diff
        return type(self).no_context(new_value, new_diff, False)

    def maximum(
        self,
        floor: Real | Atom = 0,
        /, 
    ) -> Self:
        """
        Differenatiate maximum(self, floor)
        """
        orig_value = self.value
        orig_diff = self.diff
        floor_value = floor.value if hasattr(floor, "_is_atom") else floor
        #
        if isinstance(orig_value, Real) and isinstance(orig_diff, Real):
            orig_value = _np.array(orig_value, dtype=float, )
            orig_diff = _np.array(orig_diff, dtype=float, )
        #
        new_value = _np.copy(orig_value)
        inx_floor = orig_value == floor_value
        inx_below = orig_value < floor_value
        inx_above = orig_value > floor_value
        #
        if isinstance(floor_value, Real):
            new_value[inx_below] = floor_value
        else:
            new_value[inx_below] = floor_value[inx_below]
        #
        multiplier = _np.copy(orig_value)
        multiplier[inx_floor] = 0.5
        multiplier[inx_above] = 1
        multiplier[inx_below] = 0
        new_diff = orig_diff * multiplier
        #
        return type(self).no_context(new_value, new_diff, False, )

    def mininum(
        self,
        ceiling: Real = 0,
        /,
    ) -> Self:
        """
        Differenatiate minimum(self, ceiling)
        """
        return (-self).maximum(-ceiling)
    #]


class Context:
    """
    """
    #[

    def __init__(
        self,
        atom_factory: AtomFactoryProtocol,
        equations: Iterable[_equations.Equation],
        /,
        eid_to_wrts: dict[int, tuple[Any, ...]],
        qid_to_logly: dict[int, bool] | None,
        context: dict | None,
    ) -> Self:
        """
        """
        self._eid_to_wrts = eid_to_wrts
        self._qid_to_logly = qid_to_logly or {}
        self._atom_factory = atom_factory
        #
        def get_diff_shape_for_equantion(equation, /, ) -> tuple[int, int]:
            return self._atom_factory.get_diff_shape(self._eid_to_wrts[equation.id], )
        #
        self._equations = tuple(
            _adapt_equation_for_aldi(e, get_diff_shape_for_equantion(e, ), )
            for e in equations
        )
        self._populate_atom_dict()
        #
        context = { 
            k: af_.finite_differentiator(v) 
            for k, v in context.items()
        } if context else None
        context = _adaptations.add_function_adaptations_to_context(context)
        context["Atom"] = Atom
        #
        self._equator = _equators.PlainEquator(
            self._equations,
            context=context,
        )

    def _get_num_wrts(self, eid, /, ) -> int:
        """
        """
        return len(self._eid_to_wrts[eid])

    def _populate_atom_dict(self, /, ) -> None:
        """
        """
        self._x = {}
        for eqn in self._equations:
            self._x[eqn.id] = {}
            for tok in eqn.incidence:
                atom = Atom.in_context(
                    diff=self._atom_factory.create_diff_for_token(tok, self._eid_to_wrts[eqn.id], ),
                    data_index=self._atom_factory.create_data_index_for_token(tok, ),
                    logly=self._qid_to_logly.get(tok.qid, False, ),
                )
                self._x[eqn.id][(tok.qid, tok.shift)] = atom

    def eval(
        self,
        data_array: _np.ndarray,
        column_offset: int = 0,
    ) -> Iterable[Atom]:
        """
        Evaluate and return the list of final atoms, one atom for each equation
        """
        Atom._data_context = data_array
        Atom._column_offset = column_offset
        output = self._equator.eval(self._x, 0, )
        Atom._data_context = None
        Atom._column_offset = None
        return output

    def eval_to_arrays(
        self,
        *args,
        **kwargs,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        Evaluate and return arrays of diffs and values extracted from final atoms
        """
        output = self.eval(*args, **kwargs, )
        diff = _np.vstack([x.diff for x in output])
        value = _np.vstack([x.value for x in output])
        return diff, value

    def eval_diff_to_array(
        self,
        *args,
        **kwargs,
    ) -> _np.array:
        """
        Evaluate and return array of diffs
        """
        return _np.vstack([
            x.diff for x in self.eval(*args, **kwargs, )
            if hasattr(x, "diff")
        ])

    #]


def _adapt_equation_for_aldi(
    equation: _eq.Equation,
    diff_shape: tuple[int, int],
    /,
) -> _equations.Equation:
    """
    """
    #[
    aldi_equation = equation.copy()
    aldi_equation.xtring = aldi_equation.xtring.replace("x[", f"x[{aldi_equation.id}][")
    aldi_equation.xtring += f" + Atom.zero({diff_shape}, )"
    return aldi_equation
    #]


class AtomFactoryProtocol(Protocol, ):
    """
    Protocol for creating atoms representing tokens
    """
    #[

    def create_data_index_for_token (token: Token, /, ) -> tuple[int, int | slice]: ...
    def create_diff_for_token(token: Token, wrts: tuple[Any, ...], /, ) -> _np.ndarray | int: ...
    def get_diff_shape(wrts: tuple[Any, ...], /, ) -> tuple[int, int]: ...

    #]




"""
Evaluators for dynamic period-by-period systems
"""


#[

from __future__ import annotations

import numpy as _np
from collections import namedtuple
from neqs import IterPrinter

from .. import equations as _equations
from .. import quantities as _quantities
from ._equators import Equator
from ._jacobians import Jacobian

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, Callable
    from collections.abc import Iterable

#]


_Evaluator = namedtuple(
    "_Evaluator",
    ("evaluate", "update", "get_init_guess", "iter_printer", ),
)


def create_evaluator(
    wrt_qids: Iterable[int],
    wrt_equations: Iterable[_equations.Equation],
    all_quantities: Iterable[_quantities.Quantity],
    terminal_simulator: Callable | None,
    context: dict | None,
    iter_printer_settings: dict[str, Any] | None,
) -> tuple[Callable, Callable]:
    """
    """
    wrt_qids = list(wrt_qids, )
    qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
    qid_to_name = _quantities.create_qid_to_name(all_quantities, )
    #
    # Index of loglies within wrt_qids; needs to be list not tuple
    # because of numpy indexing
    index_logly = \
        list(_quantities.generate_where_logly(wrt_qids, qid_to_logly, ))

    equator = Equator(
        wrt_equations,
        context=context,
    )

    jacobian = Jacobian(
        wrt_equations,
        wrt_qids,
        qid_to_logly,
        context=context,
    )

    iter_printer = IterPrinter(
        equation=wrt_equations,
        qids=wrt_qids,
        qid_to_logly=qid_to_logly,
        qid_to_name=qid_to_name,
        **(iter_printer_settings or {}),
    )

    def get_init_guess(
        data_array: _np.ndarray,
        columns: int | _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        maybelog_guess = data_array[wrt_qids, columns]
        maybelog_guess[index_logly] = _np.log(maybelog_guess[index_logly])
        return maybelog_guess

    def update(
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
        columns: int | _np.ndarray,
    ) -> None:
        """
        """
        #
        # Create a copy of maybelog_guess to avoid modifying the original
        # array in scipy.optimize.root
        maybelog_guess = _np.copy(maybelog_guess, )
        maybelog_guess[index_logly] = \
            _np.exp(maybelog_guess[index_logly])
        data_array[wrt_qids, columns] = maybelog_guess

    def evaluate(
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
        columns: int | _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        update(maybelog_guess, data_array, columns, )
        if terminal_simulator:
            terminal_simulator(data_array, columns, )
        equator_outcome = equator.eval(data_array, columns, )
        jacobian_outcome = jacobian.eval(data_array, columns, )
        iter_printer.next(maybelog_guess, equator_outcome, jacobian_calculated=True, )
        return equator_outcome, jacobian_outcome,

    return _Evaluator(evaluate, update, get_init_guess, iter_printer, )

    #]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import textwrap as _tw

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from numbers import Real
    from .main import Series, AxisType

#]


_NANABLE_FUNCTIONS = (
    "sum",
    "prod",
    "mean",
    "median",
    "std",
    "var",
    "max",
    "min",
    "percentile",
    "quantile",
)

_NAN_FUNCTIONS = tuple( f"nan{n}" for n in _NANABLE_FUNCTIONS )

__all__ = _NANABLE_FUNCTIONS + _NAN_FUNCTIONS


class Inlay:
    """
    """
    #[
    for n in __all__:
        code = f"""
            def {n}(
                self: Self,
                *args,
                **kwargs,
            ) -> None:
                r'''
                Function {n}
                '''
                num_periods = self.data.shape[0]
                self.data = _np.{n}(self.data, *args, axis=1, **kwargs, ).T.reshape(num_periods, -1, )
                self.trim()
        """
        exec(_tw.dedent(code, ), )
    #]



for n in __all__:
    code = f"""
        def {n}(
            self: Series,
            *args,
            axis: AxisType = 1,
            unpack_singleton: bool = True,
            **kwargs,
        ) -> Real | list[Real] | Series:
            if axis == 0:
                result = _np.{n}(self.data, *args, axis=0, **kwargs, ).T.tolist()
                if unpack_singleton and len(result) == 1:
                    result = result[0]
                return result
            elif axis == 1:
                new = self.copy()
                new.{n}(*args, **kwargs, )
                return new
            else:
                raise ValueError(f'Axis must be 0 or 1; got {{axis}}')
    """
    exec(_tw.dedent(code, ), )




"""
Slatable protocol
"""


#[
from __future__ import annotations
#]


class Slatable:
    """
    """
    #[

    __slots__ = (
        #
        "shocks_from_data",
        "stds_from_data",
        #
        "max_lag",
        "max_lead",
        #
        "databox_names",
        "databox_validators",
        "output_names",
        #
        "fallbacks",
        "overwrites",
        "qid_to_logly",
    )

    def __init__(
        self,
        *,
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
        parameters_from_data: bool = False,
        **kwargs,
    ) -> None:
        r"""
        """
        self.shocks_from_data = shocks_from_data
        self.stds_from_data = stds_from_data
        self.parameters_from_data = parameters_from_data
        #
        self.max_lag = None
        self.max_lead = None
        self.databox_names = None
        self.databox_validators = None
        self.fallbacks = None
        self.overwrites = None
        self.qid_to_logly = None
        self.output_names = None

    #]




"""
"""


#[

from __future__ import absolute_import

from typing import Literal, Iterable, Sequence
import plotly.graph_objects as _pg
import plotly.subplots as _ps
import plotly.io as _pi
import documark as _dm

from .dates import Period, Frequency

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass
#]


__all__ = (
    "plotly", "set_default_plotly_renderer",
    "make_subplots"
)


PlotlyDateAxisModeType = Literal["instant", "period", ]


_DATE_FORMAT_STYLES = {
    "sdmx": {
        Frequency.YEARLY: "%Y",
        Frequency.HALFYEARLY: "%Y-%m",
        Frequency.QUARTERLY: "%Y-Q%q",
        Frequency.MONTHLY: "%Y-%m",
        Frequency.WEEKLY: "%Y-W%W",
        Frequency.DAILY: "%Y-%m-%d",
        Frequency.INTEGER: None,
    },
    "compact": {
        Frequency.YEARLY: "%yY",
        Frequency.HALFYEARLY: "%yM%m",
        Frequency.QUARTERLY: "%yQ%q",
        Frequency.MONTHLY: "%yM%m",
        Frequency.WEEKLY: "%yW%W",
        Frequency.DAILY: "%y%b%d",
        Frequency.INTEGER: None,
    },
}


def resolve_date_format(
    date_format_style: str | dict[int, str] | None,
    frequency: Frequency,
) -> str:
    """
    """
    if not date_format_style:
        date_format_style = "sdmx"
    if isinstance(date_format_style, str):
        date_format_style = _DATE_FORMAT_STYLES[date_format_style]
    return date_format_style[frequency]


_EMPTY_SUBPLOT_TITLE = " "


def set_default_plotly_renderer(renderer: str, ) -> None:
    """
    """
    _pi.renderers.default = renderer


_DEFAULT_VERTICAL_SPACING = 0.1
_DEFAULT_HORIZONTAL_SPACING = 0.05

@_dm.reference(category="arrange", )
def make_subplots(
    rows_columns: Iterable[int] | None = None,
    subplot_titles: Sequence[str] | bool | None = True,
    vertical_spacing: float | None = None,
    horizontal_spacing: float | None = None,
    figure_height: int | None = None,
    **kwargs,
) -> _pg.Figure:
    """
················································································

==Create a figure with multiple subplots==

················································································
    """
    if vertical_spacing is None:
        vertical_spacing = _DEFAULT_VERTICAL_SPACING
    if horizontal_spacing is None:
        horizontal_spacing = _DEFAULT_HORIZONTAL_SPACING
    num_rows = rows_columns[0] if rows_columns else 1
    num_columns = rows_columns[1] if rows_columns else 1
    total_num_subplots = num_rows * num_columns
    if subplot_titles is True:
        subplot_titles = (_EMPTY_SUBPLOT_TITLE, )*total_num_subplots
    figure = _ps.make_subplots(
        rows=num_rows,
        cols=num_columns,
        subplot_titles=subplot_titles or None,
        vertical_spacing=vertical_spacing,
        horizontal_spacing=horizontal_spacing,
        **kwargs,
    )
    if figure_height is not None:
        figure.update_layout(height=figure_height, )
    return figure

@_dm.reference(category="custom", )
def highlight(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
    subplot_ref: tuple[int, int] | int | None = None,
    color: str = "rgba(0, 0, 0, 0.15)",
) -> _pg.Figure:
    """
················································································

==Highlight a certain date span in charts==

Highlight a certain time span in a figure. The span can be specified either as a
time `Span`, or a tuple of two `Periods`.

    irispie.plotly.highlight(span)

················································································
    """
    highlight_start, highlight_end = _resolve_highlight_span(figure, span, )
    _, index = resolve_subplot_reference(figure, subplot_ref, )
    xref = f"x{index+1}" if index else "x"
    yref = f"y{index+1} domain" if index else "y domain"
    shape = {
        "type": "rect",
        "xref": xref,
        "x0": highlight_start,
        "x1": highlight_end,
        "yref": yref,
        "y0": 0,
        "y1": 1,
        "fillcolor": color,
        "line": {"width": 0, },
    }
    figure.add_shape(shape, )


def vline(
    figure,
    vline_period,
    subplot_ref: tuple[int, int] | int | None = None,
    color: str = "rgba(0, 0, 0, 0.5)",
    width: int = None,
) -> None:
    r"""
    """
    date_axis_mode = get_date_axis_mode(figure, )
    row_column, _ = resolve_subplot_reference(figure, subplot_ref, )
    if hasattr(vline_period, "to_plotly_date"):
        vline_period = vline_period.to_plotly_date(mode=date_axis_mode, )
    figure.add_vline(vline_period, **row_column, line={"color": color, "width": width, }, )


def freeze_span(
    figure: _pg.Figure,
    span: Iterable[Period],
    /,
    subplot: tuple[int, int] | int | None = None,
) -> None:
    """
    """
    start, end = _resolve_figure_span(figure, span, )
    xaxis_update = {
        "range": (start, end),
        "autorange": False,
    }
    row_column, *_ = resolve_subplot_reference(figure, subplot, )
    figure.update_xaxes(xaxis_update, **row_column, )


def _resolve_figure_span(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
) -> tuple[str, str]:
    """
    """
    #[
    if hasattr(span, "start", ) and hasattr(span, "end", ):
        start = span.start if span.start else None
        end = span.end if span.end else None
    elif span:
        span = tuple(span, )
        start, end = span[0], span[-1]
    else:
        start, end = None, None
    figure_span = figure.layout.xaxis.range
    mode = get_date_axis_mode(figure, )
    start = start.to_plotly_edge_before(mode=mode, ) if start is not None else figure_span[0]
    end = end.to_plotly_edge_after(mode=mode, ) if end is not None else figure_span[-1]
    return start, end,
    #]


def _resolve_highlight_span(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
) -> tuple[str, str]:
    """
    """
    #[
    if hasattr(span, "start", ) and hasattr(span, "end", ):
        start = span.start if span.start else None
        end = span.end if span.end else None
    elif span:
        span = tuple(span, )
        start, end = span[0], span[-1]
    else:
        start, end = None, None
    figure_span = figure.layout.xaxis.range
    mode = get_date_axis_mode(figure, )
    highlight_start = start.to_plotly_edge_before(mode=mode, ) if start is not None else figure_span[0]
    highlight_end = end.to_plotly_edge_after(mode=mode, ) if end is not None else figure_span[-1]
    return highlight_start, highlight_end,
    #]


def resolve_subplot_reference(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
) -> tuple[dict[str, int], int]:
    r"""
    """
    if subplot is None:
        return {}, None
    rows, columns = figure._get_subplot_rows_columns()
    num_rows = len(rows)
    num_columns = len(columns)
    num_tiles = num_rows * num_columns
    if isinstance(subplot, Sequence):
        row = subplot[0]
        column = subplot[1]
        index = row * num_columns + column
        row_column = {"row": row+1, "col": column+1, }
        return row_column, index,
    if isinstance(subplot, int):
        if subplot < 0:
            subplot += num_tiles
        index = subplot
        row = index // num_columns
        column = index % num_columns
        row_column = {"row": row+1, "col": column+1, }
        return row_column, index,
    raise TypeError(f"Invalid subplot type: {type(subplot)}")


def update_subplot_title(
    figure: _pg.Figure,
    subplot_ref: tuple[int, int] | int,
    title: str,
) -> None:
    r"""
    """
    row_column, index = resolve_subplot_reference(figure, subplot_ref, )
    annotation = next(figure.select_annotations(index, ), None, )
    if annotation:
        annotation.text = title


def get_date_axis_mode(figure: _pg.Figure, ) -> PlotlyDateAxisModeType | None:
    """
    """
    xaxis = figure.layout.xaxis
    return xaxis.ticklabelmode if xaxis.type == "date" else None


@_dm.reference(
    path=("visualization_reporting", "ez_plotlyper.md", ),
    categories={
        "arrange": "Arranging charts",
        "custom": "Customizing charts",
    },
)
class plotly:
    """
················································································

Plotly wrapper
===============

················································································
    """
    #[

    Figure = _pg.Figure
    make_subplots = staticmethod(make_subplots)
    update_subplot_title = staticmethod(update_subplot_title)
    resolve_subplot_reference = staticmethod(resolve_subplot_reference)
    highlight = staticmethod(highlight)
    freeze_span = staticmethod(freeze_span)

    #]




"""
# Stacked-time linear system

$$
\tilda x_t = 
$$

"""


#[

from __future__ import annotations

import numpy as _np

from ..fords import initializers as _initializers
from ..fords.solutions import Solution

from ._invariants import Invariant

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from numbers import Real

#]


mat_pow = _np.linalg.matrix_power


class Variant:
    """
    """

    _system_matrix_slots = (
        "TT", "PP", "RR",
        "AA", "BB", "CC", "HH",
    )

    _system_intercept_slots = (
        "KK", "DD",
    )

    __slots__ = (
        *_system_matrix_slots,
        *_system_intercept_slots,
        "init_med",
        "init_mse",
        "Xi",
        "MM",
        "std_name_to_value",
        "U",
        "has_unknown_initial",
    )

    def __init__(self, ) -> None:
        """
        """
        for n in self.__slots__:
            setattr(self, n, None, )

    @classmethod
    def from_solution_and_stds(
        klass,
        invariant: Invariant,
        solution: Solution,
        std_name_to_value: dict[str, Real],
        cov_u: _np.ndarray,
    ) -> Self:
        """
        """
        self = klass()
        #
        num_periods = invariant.num_periods
        forward = num_periods - 1
        T, P, _, K, Z, H, D, U = solution.unpack_triangular_solution()
        R = solution.expand_triangular_solution(forward=forward, )
        index_xi = invariant.index_xi
        index_u = invariant.index_u
        index_v = invariant.index_v
        index_y = invariant.index_y
        index_w = invariant.index_w
        num_xi = T.shape[0]
        num_u = P.shape[1]
        num_v = R[0].shape[1]
        num_y_included = len(index_y)
        num_u_included = len(index_u)
        num_v_included = len(index_v)
        num_w_included = len(index_w)
        Px = _np.hstack([P[:, index_u], _np.zeros((num_xi, num_u_included*(num_periods-1)))], )
        Rx = _np.hstack([r[:, index_v] for r in R])
        Hx = _np.hstack([H[index_y, :][:, index_w], _np.zeros((num_y_included, num_w_included*(num_periods-1)))], )
        U = U[index_xi, :]
        Z = Z[index_y, :]
        D = D[index_y]
        #
        for n in self._system_matrix_slots:
            setattr(self, n, [], )
        for n in self._system_intercept_slots:
            setattr(self, n, [], )
        #
        full_TT = _np.eye(num_xi, )
        full_KK = _np.zeros((num_xi, ), )
        full_PP = _np.zeros((num_xi, num_u_included*num_periods, ), )
        full_RR = _np.zeros((num_xi, num_v_included*num_periods, ), )
        for k in range(num_periods):
            full_TT = T @ full_TT
            full_KK = T @ full_KK + K
            full_PP = T @ full_PP + Px
            full_RR = T @ full_RR + Rx
            #
            self.TT.append(U @ full_TT)
            self.KK.append(U @ full_KK)
            self.PP.append(U @ full_PP)
            self.RR.append(U @ full_RR)
            #
            self.AA.append(Z @ full_TT)
            self.BB.append(Z @ full_PP)
            self.CC.append(Z @ full_RR)
            self.DD.append(Z @ full_KK + D)
            self.HH.append(Hx)
            #
            Px = _move_forward(Px, num_u_included)
            Rx = _move_forward(Rx, num_v_included)
            Hx = _move_forward(Hx, num_w_included)
        #
        for n in self._system_matrix_slots:
            a = getattr(self, n)
            setattr(self, n, _np.vstack(a, ), )
        for n in self._system_intercept_slots:
            a = getattr(self, n)
            setattr(self, n, _np.hstack(a, ), )
        self.U = U
        #
        self.std_name_to_value = std_name_to_value
        #
        self.init_med, self.init_mse, Xi = \
            _initializers.initialize(solution, cov_u, )
        #
        self.has_unknown_initial = Xi is not None
        if self.has_unknown_initial:
            self.Xi = Xi
            self.MM = self.AA @ Xi
        #
        return self


def _move_forward(X: _np.ndarray, by: int, /, ) -> _np.ndarray:
    num_rows = X.shape[0]
    return _np.hstack([_np.zeros((num_rows, by)), X[:, :-by], ], )




"""
Vector autoregression (VAR) module
"""

from .red_vars import *
from .red_vars import __all__ as red_vars_all

from .str_vars import *
from .str_vars import __all__ as str_vars_all

from .prior_dummies import *
from .prior_dummies import __all__ as prior_dummies_all

__all__ = (
    *red_vars_all,
    *str_vars_all,
    *prior_dummies_all,
)




r"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
from typing import Callable
from numbers import Real

from . import solutions as _solutions

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass

#]


__all__ = (
    "std_from_cov",
    "cov_from_std",
    "CovarianceSimulator",
)


def get_autocov_square(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    cov_w: _np.ndarray,
    order: int,
) -> _np.ndarray:
    """
    Autocovariance of [xi; y] with nans for unstable elements
    """
    #[
    boolex_stable = _np.hstack((
        solution.boolex_stable_transition_vector,
        solution.boolex_stable_measurement_vector,
    ))
    def fill_nans(cov: _np.ndarray, ) -> _np.ndarray:
        cov[~boolex_stable, :] = _np.nan
        cov[:, ~boolex_stable] = _np.nan
        return cov
    return tuple(
        fill_nans(cov, )
        for cov in get_autocov_square_00(solution, cov_u, cov_w, order, )
    )
    #]


def get_autocov_square_00(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    cov_w: _np.ndarray,
    order: int,
) -> tuple[_np.ndarray, ...]:
    """
    Autocovariance (up to order) of [xi; y] based on stable alpha and a zero cov assumption for unstable alpha
    """
    #[
    num_alpha = solution.num_alpha
    Ua = solution.Ua
    #
    def _transform_cov_triangular_to_square(cov: _np.ndarray, ) -> _np.ndarray:
        cov[:num_alpha, :] = Ua @ cov[:num_alpha, :]
        cov[:, :num_alpha] = cov[:, :num_alpha] @ Ua.T
        return cov
    #
    return tuple(
        _transform_cov_triangular_to_square(cov, )
        for cov in get_autocov_triangular_00(solution, cov_u, cov_w, order, )
    )
    #]


def get_autocov_triangular_00(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    cov_w: _np.ndarray,
    order: int,
) -> tuple[_np.ndarray, ...]:
    """
    Autocovariance (up to order) of [alpha; y] based on stable alpha
    """
    #[
    cov_triangular_00 = get_cov_triangular_00(solution, cov_u, cov_w, )
    #
    num_unit_roots = solution.num_unit_roots
    num_alpha = solution.num_alpha
    num_y = solution.num_y
    Ta_00 = solution.Ta.copy()
    Ta_00[:num_unit_roots, :] = 0
    Ta_00[:, :num_unit_roots] = 0
    Za = solution.Za
    A = _np.block([
        [Ta_00, _np.zeros((num_alpha, num_y), dtype=float, )],
        [Za @ Ta_00, _np.zeros((num_y, num_y), dtype=float, )],
    ])
    #
    autocov_triangular_00 = [None, ] * (order + 1)
    autocov_triangular_00[0] = cov_triangular_00
    for i in range(order):
        autocov_triangular_00[i+1] = A @ autocov_triangular_00[i]
    return tuple(autocov_triangular_00)
    #]


def get_cov_triangular_00(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    cov_w: _np.ndarray,
) -> _np.ndarray:
    """
    Covariance of [alpha; y] based on stable alpha
    """
    #[
    # Triangular transition equations
    #
    cov_alpha_00 = get_cov_alpha_00(solution, cov_u, )
    #
    # Measurement equations
    #
    num_unit_roots = solution.num_unit_roots
    cov_alpha_stable = cov_alpha_00[num_unit_roots:, num_unit_roots:]
    Za_stable = solution.Za_stable
    H = solution.H
    sigma_w = H @ cov_w @ H.T
    cov_y_00 = Za_stable @ cov_alpha_stable @ Za_stable.T + sigma_w
    #
    # Cross-covariance between transition and measurement equations
    #
    Za = solution.Za
    cov_alpha_y_00 = cov_alpha_00 @ Za.T
    #
    # Full covariance matrix based on stable alpha
    #
    return _np.block([
        [cov_alpha_00, cov_alpha_y_00],
        [cov_alpha_y_00.T, cov_y_00],
    ])
    #]


def get_cov_alpha_00(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
) -> _np.ndarray:
    """
    Covariance of stable alpha, and full alpha based on stable alpha
    """
    #[
    num_unit_roots = solution.num_unit_roots
    Ta_stable = solution.Ta_stable
    Pa_stable = solution.Pa_stable
    sigma_u = Pa_stable @ cov_u @ Pa_stable.T
    cov_alpha_stable = _sp.linalg.solve_discrete_lyapunov(Ta_stable, sigma_u, )
    cov_alpha_00 = _np.zeros(solution.Ta.shape, dtype=float, )
    cov_alpha_00[num_unit_roots:, num_unit_roots:] = cov_alpha_stable
    return cov_alpha_00
    #]


def acorr_from_acov(
    acov_by_order: tuple[_np.ndarray, ...],
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    acorr_by_order = [None, ] * len(acov_by_order)
    #
    # Scale matrix has inv_std[i]*inv_std[j]
    scale_matrix = _get_scale_matrix(acov_by_order[0], )
    #
    # Rescale autocovariance matrices to autocorrelation matrices order by
    # order
    acorr_by_order = tuple(
        i * scale_matrix
        for i in acov_by_order
    )
    #
    return acorr_by_order
    #]


def _get_scale_matrix(
    cov: _np.ndarray,
    tolerance: float = 1e-12,
) -> _np.ndarray:
    """
    """
    #[
    inv_std = _np.diag(cov).copy()
    index_positive = _np.abs(inv_std) > 0
    inv_std[index_positive] = 1 / _np.sqrt(inv_std[index_positive])
    inv_std[~index_positive] = 0
    return inv_std.reshape(-1, 1, ) @ inv_std.reshape(1, -1, )
    #]


def _is_singleton(
    array: _np.ndarray,
) -> bool:
    """
    """
    return array.ndim == 2


def _add_dim_to_singleton(
    array_by_order: tuple[_np.ndarray, ...],
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    return tuple(
        array[:, :, _np.newaxis]
        for array in array_by_order
    )
    #]


def _remove_dim_from_singleton(
    array_by_order: tuple[_np.ndarray, ...],
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    return tuple(
        array[:, :, 0]
        for array in array_by_order
    )
    #]


def _lyapunov(
    T: _np.ndarray,
    Sigma: _np.ndarray,
) -> _np.ndarray:
    """
    """

    def left_div(A: _np.ndarray, B: _np.ndarray) -> _np.ndarray:
        """
        Solve A / B
        """
        return _np.linalg.solve(A, B, )

    def right_div(B: _np.ndarray, A: _np.ndarray) -> _np.ndarray:
        """
        Solve B / A
        """
        return _np.linalg.solve(A.T, B.T, ).T

    C = _np.zeros_like(T)
    i = T.shape[0]
    Tt = T.T

    while i >= 1:
        if i == 1 or T[i-1, i-2] == 0:
            # 1x1 block with a real eigenvalue.
            C[i-1, i:] = C[i:, i-1].T
            c = right_div(
                Sigma[i-1, :i] + T[i-1, i-1] * C[i-1, i:] @ Tt[i:, :i] + T[i-1, i:] @ C[i:, :] @ Tt[:, :i],
                _np.eye(i) - T[i-1, i-1] * Tt[:i, :i],
            )
            C[i-1, :i] = c
            i -= 1
        else:
            # 2x2 block corresponding to a pair of complex eigenvalues.
            C[i-2:i, i:] = C[i:, i-2:i].T
            X = T[i-2:i, i-2:i] @ C[i-2:i, i:] @ Tt[i:, :i] + T[i-2:i, i:] @ C[i:, :] @ Tt[:, :i] + Sigma[i-2:i, :i]
            # Solve
            #     c = T(i-1:i, i-1:i)*c*Tt(1:i, 1:i)
            # Transpose the equation first
            #     c' = Tt'*c'*T' + X'
            # so that the below kronecker product becomes faster to evaluate,
            # then vectorize
            #     vec(c') = kron(T, Tt')*vec(c') + vec(X').
            Xt = X.T
            U = Tt[:i, :i].T
            k = _np.block([
                [ T[i-2, i-2]*U, T[i-2, i]*U ],
                [ T[i-1, i-2]*U, T[i-1, i]*U ],
            ])
            ct = left_div(_np.eye(2*i) - k, Xt.ravel(), )
            C[i-2:i, :i] = ct.reshape(i, 2).T
            i -= 2

    return C


def symmetrize(X: _np.ndarray, ) -> _np.ndarray:
    """
    """
    return (X + X.T) / 2


def std_from_cov(
    cov: _np.ndarray,
    trim_negative: bool = True,
) -> _np.ndarray:
    r"""
    """
    #[
    diag = _np.diag(cov, )
    if trim_negative:
        diag = _np.maximum(diag, 0)
    return _np.sqrt(diag, )
    #]


def cov_from_std(
    std: _np.ndarray,
    trim_negative: bool = True,
) -> _np.ndarray:
    r"""
    """
    #[
    if trim_negative:
        std = _np.maximum(std, 0)
    return _np.diag(std**2, )
    #]


_DEFAULT_NUMPY_RNG = _np.random.default_rng()
_DEFAULT_GENERATOR = _DEFAULT_NUMPY_RNG.standard_normal


class CovarianceSimulator:
    r"""
    """
    #[

    def __init__(
        self,
        cov: _np.ndarray,
        mean: _np.ndarray | Real | None = None,
        #
        generator: Callable | None = None,
        make_symmetric: bool = True,
        trim_negative: bool = True,
    ) -> None:
        r"""
        """
        if make_symmetric:
            cov = symmetrize(cov, )
        eig_values, eig_vectors = _np.linalg.eigh(cov, )
        if trim_negative:
            eig_values = _np.maximum(eig_values, 0)
        std_devs = _np.sqrt(eig_values, )
        self.factor = eig_vectors @ _np.diag(std_devs, )
        self._populate_mean(mean, )
        self.generator = generator if generator is not None else _DEFAULT_GENERATOR

    def _populate_mean(
        self,
        mean: _np.ndarray | Real | None,
    ) -> None:
        r"""
        """
        shape = (self.num_variables, 1, )
        if mean is None:
            mean = 0
        if isinstance(mean, Real):
            mean = _np.full(shape, mean, dtype=float, )
        self.mean = mean.reshape(shape, order="F", )

    @property
    def num_variables(self, /, ) -> int:
        r"""
        """
        return self.factor.shape[0]

    def simulate(
        self,
        num_draws: int,
    ) -> _np.ndarray:
        r"""
        """
        shape = self.num_variables, num_draws,
        x = self.factor @ self.generator(shape, )
        if self.mean is not None:
            x += self.mean
        return x

    #]




"""
Evaluators for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np
from types import (SimpleNamespace, )

from .. import equations as _equations
from .. import quantities as _quantities
from ..aldi.maps import (ArrayMap, )
from ..incidences.main import (Token, )
from ._equators import (Equator, )
from ._jacobians import (Jacobian, )
from ..evaluators.base import (Evaluator, )

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, Callable, )
    from collections.abc import (Iterable, Sequence, )
    from ..fords.terminators import (Terminator, )
#]


def _create_update_map(
    wrt_spots: Iterable[Token],
) -> ArrayMap:
    """
Create a map for updating the data array with the new guess.
    """
    update_map = ArrayMap()
    lhs_rows, lhs_columns = zip(*wrt_spots, )
    lhs_rows = _np.array(lhs_rows, dtype=int, )
    lhs_columns = _np.array(lhs_columns, dtype=int, )
    update_map.lhs = (lhs_rows, lhs_columns, )
    return update_map


def create_evaluator(
    wrt_spots: Iterable[Token],
    columns_to_eval: Sequence[int],
    wrt_equations: Iterable[_equations.Equation],
    all_quantities: Iterable[_quantities.Quantity],
    terminator: Terminator | None,
    context: dict | None,
) -> SimpleNamespace:
    """
    """
    #[
    num_columns_to_eval = len(columns_to_eval)
    wrt_spots = list(wrt_spots, )
    needs_terminal = terminator is not None
    #
    qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
    qid_to_name = _quantities.create_qid_to_name(all_quantities, )
    #
    # Index of loglies within wrt_spots; needs to be list not tuple
    # because of numpy indexing
    index_logly = list(_quantities.generate_where_logly(
        ( tok.qid for tok in wrt_spots ), qid_to_logly,
    ))

    update_map = _create_update_map(wrt_spots, )

    equator = Equator(
        wrt_equations,
        columns=columns_to_eval,
        context=context,
    )

    jacobian_wrt_spots = tuple(wrt_spots)
    if needs_terminal:
        jacobian_wrt_spots += tuple(terminator.terminal_wrt_spots)

    jacobian = Jacobian(
        wrt_equations,
        jacobian_wrt_spots,
        qid_to_logly,
        sparse=True,
        context=context,
        columns_to_eval=columns_to_eval,
        terminator=terminator,
    )

    def get_init_guess(data_array: _np.ndarray, /, ) -> _np.ndarray:
        """
        """
        maybelog_guess = data_array[update_map.lhs[0], update_map.lhs[1]]
        maybelog_guess[index_logly] = _np.log(maybelog_guess[index_logly])
        return maybelog_guess

    def update(
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
    ) -> None:
        """
        """
        maybelog_guess = _np.copy(maybelog_guess, )
        maybelog_guess[index_logly] = _np.exp(maybelog_guess[index_logly])
        data_array[update_map.lhs[0], update_map.lhs[1]] = maybelog_guess

    def eval_func_jacob(
        maybelog_guess: _np.ndarray | None,
        data_array: _np.ndarray,
    ) -> tuple[_np.ndarray, _sp.sparse.csc_matrix]:
        """
        """
        if maybelog_guess is not None:
            update(maybelog_guess, data_array, )
        #
        if needs_terminal:
            terminator.terminate_simulation(data_array, )
        #
        # The equator evaluates to a tuple of 1D arrays, each array for
        # one equation across all periods. Reorganize into a 1D array ordered
        # as follows:
        #
        # [eq1[t=0], eq2[t=0], ..., eq1[t=1], eq2[t=1], ...]
        #
        equator_outcome = equator.eval(data_array, )
        equator_outcome = _np.vstack(equator_outcome, ).flatten(order="Fortran"[0], )
        #
        jacobian_outcome = jacobian.eval(data_array, )
        if needs_terminal:
            jacobian_outcome = terminator.terminate_jacobian(jacobian_outcome, )
        return equator_outcome, jacobian_outcome, 

    def eval_func(
        maybelog_guess: _np.ndarray | None,
        data_array: _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        if maybelog_guess is not None:
            update(maybelog_guess, data_array, )
        if needs_terminal:
            terminator.terminate_simulation(data_array, )
        equator_outcome = equator.eval(data_array, )
        return _np.vstack(equator_outcome, ).flatten(order="F", )

    def eval_jacob(
        maybelog_guess: _np.ndarray | None,
        data_array: _np.ndarray,
    ) -> _sp.sparse.csc_matrix:
        """
        """
        if maybelog_guess is not None:
            update(maybelog_guess, data_array, )
        if needs_terminal:
            terminator.terminate_simulation(data_array, )
        jacobian_outcome = jacobian.eval(data_array, )
        if needs_terminal:
            jacobian_outcome = terminator.terminate_jacobian(jacobian_outcome, )
        return jacobian_outcome

    return Evaluator(
        eval_func_jacob=eval_func_jacob,
        eval_func=eval_func,
        eval_jacob=eval_jacob,
        update=update,
        get_init_guess=get_init_guess,
    )

    #]





"""
Merge inlay
"""


#[

from __future__ import annotations

from typing import (TYPE_CHECKING, )
import warnings as _wa
import documark as _dm

from .. import wrongdoings as _wrongdoings
from ..series import main as _series
from . import main as _databoxes

if TYPE_CHECKING:
    from typing import (Self, Iterable, Literal, )
    MergeStrategyType = Literal[
        "stack", # Stack values as variants
        "hstack", # Legacy alias for stack, do not include in the docstring
        "replace", # Replace the existing value with the new one
        "discard", # Discard the new value and keep the existing one
        "silent", # Exactly the same as discard
        "warning", # Same as discard with a warning
        "error", # Throw an error for the first duplicate key
        "critical", # Throw a critical error for the first duplicate key
    ]

#]


class Inlay:
    """
    """
    #[

    @classmethod
    def merged(
        klass,
        databoxes: Iterable[Self],
        merge_strategy: MergeStrategyType = "stack",
    ) -> Self:
        """
        """
        out = klass()
        out.merge(databoxes, merge_strategy, )
        return out

    @_dm.reference(category="multiple", )
    def merge(
        self: Self,
        other: Self | Iterable[Self],
        merge_strategy: MergeStrategyType = "stack",
        # Do not include the following in the docstring
        action = None,
    ) -> None:
        r"""
................................................................................

==Merge Databoxes==

Combine one or more databoxes into a single databox using a specified merge
strategy to handle potential conflicts between duplicate keys.

    self.merge(
        other,
        merge_strategy="stack",
    )


### Input arguments ###

???+ input "other"
    The databox or iterable of databoxes to merge into the current databox. If
    merging a single databox, it should be passed directly; for multiple
    databoxes, pass an iterable containing all.

???+ input "merge_strategy"
    Determines how to process keys that exist in more than one databox. The
    default strategy is `"stack"`.

    * `"stack"`: Stack values; this means combine time series into multiple
    columns, or combine lists, or convert non-lists to lists for stacking.

    * `"replace"`: Replace existing values with new values.

    * `"discard"` and `"silent"`: Retain original values and ignore new values.

    * `"warning"`: Behave like `"discard"` but issue a warning for each conflict.

    * `"error"`: Raise an error on encountering the first duplicate key.

    * `"critical"`: Raise a critical error on encountering the first duplicate key.


### Returns ###

    This method modifies the databox in place and returns `None`.

................................................................................
        """
        # Legacy name
        if action is not None:
            _wa.warn("The 'action' input argument is deprecated; use 'merge_strategy' instead", )
            merge_strategy = action
        #
        merge_strategy_func = _MERGE_STRATEGY_DISPATCH[merge_strategy]
        stream = _wrongdoings.create_stream(
            merge_strategy,
            "Duplicate keys when merging databoxes",
            when_no_stream="silent",
        )
        if hasattr(other, "items", ):
            other = (other, )
        for t in other:
            for key, value in t.items():
                if key in self:
                    merge_strategy_func(self, key, value, stream, )
                else:
                    self[key] = value
        stream._raise()

    #]


def _merge_stack(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    Horizontal stack of values: time series are concatenated, lists are
    extended, non-lists are converted to lists and extended.
    """
    #[
    if isinstance(value, _series.Series, ):
        self[key] = self[key] | value
        return
    if not isinstance(self[key], list):
        self[key] = [self[key], ]
    if not isinstance(value, list):
        value = [value, ]
    self[key] += value
    #]


def _merge_replace(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    self[key] = value
    #]


def _merge_discard(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    pass
    #]


def _merge_report(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    stream.add(key, )
    #]


_MERGE_STRATEGY_DISPATCH = {
    "stack": _merge_stack,
    "hstack": _merge_stack,
    "replace": _merge_replace,
    "discard": _merge_discard,
    "silent": _merge_report,
    "warning": _merge_report,
    "error": _merge_report,
    "critical": _merge_report,
}




"""
Data arrays with row and column names
"""


#[
from __future__ import annotations

from typing import (Self, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, Iterator, )
import numpy as _np
import numpy as _np
import functools as _ft
import itertools as _it

from ..series.main import (Series, )
from ..databoxes.main import (Databox, )
from ..incidences import main as _incidences
from ..conveniences import iterators as _iterators
from ..dates import (Period, Span, )
from .. import dates as _dates
from .. import has_variants as _has_variants

from . import _invariants as _invariants
from . import _variants as _variants
#]


__all__ = (
    "Dataslate",
)


class SlatableProtocol(Protocol, ):
    """
    """
    max_lag: int
    max_lead: int
    databox_names: Iterable[str]
    databox_validators: dict[str, Callable]
    fallbacks: dict[str, Real]
    overwrites: dict[str, Real]
    output_names: Iterable[str]
    qid_to_logly: dict[int, bool | None]


class Dataslate(
    _has_variants.Mixin,
):
    """
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    def __init__(self, /, ) -> None:
        self._invariant = None
        self._variants = []

    @classmethod
    def skeleton(
        klass,
        other: Self,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = other._invariant
        self._variants = []
        return self

    @classmethod
    def nan_from_names_periods(
        klass,
        names: Iterable[str],
        periods: Iterable[Period] | string,
        /,
        num_variants: int = 1,
        **kwargs,
    ) -> Iterator[Self]:
        """
        """
        names = tuple(names or databox.keys())
        periods = _dates.ensure_period_tuple(periods, )
        num_names = len(names)
        num_periods = len(periods)
        self = klass()
        self._invariant = _invariants.Invariant(names, periods, **kwargs, )
        self._variants = [
            _variants.Variant.nan_data_array(self._invariant, )
            for _ in range(num_variants, )
        ]
        return self

    @classmethod
    def nan_from_template(
        klass,
        other: Self,
        /,
        num_variants: int = 1,
    ) -> Self:
        """
        """
        self = klass.skeleton(other, )
        self._variants = [
            _variants.Variant.nan_data_array(self._invariant, )
            for _ in range(num_variants, )
        ]
        return self

    @classmethod
    def from_databox(
        klass,
        databox: Databox | dict,
        names: Iterable[str] | None,
        periods: Iterable[Period] | string,
        /,
        num_variants: int = 1,
        fallbacks: dict[str, Number] | None = None,
        overwrites: dict[str, Number] | None = None,
        clip_data_to_base_span: bool = False,
        validators: dict[str, Callable] | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        if names is None:
            names = databox.keys()
        names = tuple(names)
        periods = _dates.ensure_period_tuple(periods, )
        #
        if validators:
            Databox.validate(databox, validators, )
        #
        from_until = periods[0], periods[-1] if periods else ()
        item_iterator = \
            _ft.partial(_slate_value_variant_iterator, from_until=from_until, )
        #
        databox_variant_iterator = \
            Databox.iter_variants(databox, item_iterator=item_iterator, keys=names, )
        #
        fallbacks_variant_iterator = (
            Databox.iter_variants(fallbacks, item_iterator=item_iterator, )
            if fallbacks else _it.repeat(None, )
        )
        #
        overwrites_variant_iterator = (
            Databox.iter_variants(overwrites, item_iterator=item_iterator, )
            if overwrites else _it.repeat(None, )
        )
        #
        zipped = zip(
            range(num_variants),
            databox_variant_iterator,
            fallbacks_variant_iterator,
            overwrites_variant_iterator,
        )
        #
        self = klass()
        self._invariant = _invariants.Invariant(names, periods, **kwargs, )
        self._variants = [
            _variants.Variant.from_databox_variant(
                databox_v, self._invariant,
                fallbacks=fallbacks_v,
                overwrites=overwrites_v,
                clip_data_to_base_span=clip_data_to_base_span,
            )
            for vid, databox_v, fallbacks_v, overwrites_v in zipped
        ]
        return self

    @classmethod
    def from_databox_for_slatable(
        klass,
        slatable: SlatableProtocol,
        databox: Databox | dict,
        base_span: Iterable[Period],
        /,
        extra_databox_names: Iterable[str] | None = None,
        prepend_initial: bool = True,
        append_terminal: bool = True,
        clip_data_to_base_span: bool = False,
        **kwargs,
    ) -> Self:
        """
        """
        names = tuple(slatable.databox_names, )
        if slatable.descriptions:
            descriptions = tuple(slatable.descriptions, )
        else:
            descriptions = tuple("" for _ in names)
        if extra_databox_names:
            names = names + tuple(i for i in extra_databox_names if i not in names)
            descriptions = descriptions + tuple("" for _ in extra_databox_names)
        periods, base_columns, *min_max_shift = _get_extended_span(
            slatable, base_span,
            prepend_initial=prepend_initial,
            append_terminal=append_terminal,
        )
        #
        return klass.from_databox(
            databox, names, periods,
            base_columns=base_columns,
            descriptions=descriptions,
            min_max_shift=min_max_shift,
            qid_to_logly=slatable.qid_to_logly,
            fallbacks=slatable.fallbacks,
            overwrites=slatable.overwrites,
            clip_data_to_base_span=clip_data_to_base_span,
            output_names=slatable.output_names,
            validators=slatable.databox_validators,
            **kwargs,
        )

    @property
    def descriptions(self, /, ) -> tuple[str, ...] | None:
        return self._invariant.descriptions

    @property
    def logly_indexes(self, /, ) -> tuple[int, ...] | None:
        return self._invariant.logly_indexes

    @property
    def periods(self, /, ) -> tuple[Period]:
        """
        """
        return self._invariant.periods

    @property
    def num_base_periods(self, /, ) -> int:
        """
        """
        return self._invariant.num_base_periods

    @property
    def start(self, /, ) -> Period:
        """
        """
        return self._invariant.periods[0]

    first_period = start

    @property
    def end(self, /, ) -> Period:
        """
        """
        return self._invariant.periods[-1]

    last_period = end

    @property
    def base_periods(self, /, ) -> tuple[Period]:
        """
        """
        return self._invariant.base_periods

    @base_periods.setter
    def base_periods(self, base_periods: Iterable[Period], /, ) -> None:
        """
        """
        self._invariant.base_periods = base_periods

    @property
    def base_start(self, /, ) -> Period:
        """
        """
        return self._invariant.base_periods[0]

    @property
    def names(self, /, ) -> tuple[str]:
        """
        """
        return self._invariant.names

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self.names)

    @property
    def num_initials(self, /, ) -> int:
        """
        """
        return -self._invariant.min_max_shift[0]

    @property
    def num_terminals(self, /, ) -> int:
        """
        """
        return self._invariant.min_max_shift[1]

    @property
    def output_names(self, /, ) -> tuple[str, ...]:
        """
        """
        return tuple(self.names[i] for i in self._invariant.output_qids)

    def copy(
        self, 
        invariant=True,
        variants=True,
    ) -> Self:
        """
        """
        new = type(self)()
        if invariant:
            new._invariant = self._invariant.copy()
        else:
            new._invariant = self._invariant
        #
        if variants:
            new._variants = [i.copy() for i in self._variants]
        else:
            new._variants = [i for i in self._variants]
        #
        return new

    def rename(self, old_name_to_new_name: dict[str, str], /, ) -> None:
        """
        """
        self._invariant.names = tuple(
            old_name_to_new_name.get(i, i)
            for i in self._invariant.names
        )

    def extend(
        self: Self,
        other: Self,
        /,
    ) -> None:
        """
        """
        # TODO: Check if the other is compatible
        self._variants.extend(other._variants, )

    def nan_copy(self, /, ) -> Self:
        """
        """
        return self.nan_from_template(self, num_variants=self.num_variants, )

    def update_columns_from(
        self,
        other: Self,
        columns: Iterable[int],
        /,
    ) -> None:
        """
        """
        for self_v, other_v in zip(self._variants, other._variants, ):
            self_v.update_columns_from(other_v, columns, )

    def remove_initial_data(self, /, ) -> None:
        """
        """
        if self.num_initials:
            self.data = self.data[:, self.num_initials:]

    def to_databox(
        self,
        /,
        target_db: Databox | None = None,
        span: Literal["base", "full", ] = "full",
    ) -> Databox:
        """
        Add data from a dataslate to a new or existing databox
        """
        #[
        if target_db is None:
            target_db = Databox()
        if span == "full":
            start = self.start
            column_slice = slice(None, None, )
        elif span == "base":
            start = self.base_start
            column_slice = self.base_slice
        num_names = self.num_names
        num_variants = self.num_variants
        for qid in self._invariant.output_qids:
            name = self._invariant.names[qid]
            description = self._invariant.descriptions[qid]
            values = _np.vstack([
                v.data[qid, column_slice] for v in self._variants
            ]).T
            target_db[name] = Series._guaranteed(
                start=start,
                values=values,
                description=description,
            )
        return target_db
        #]

    for n in ["num_periods", "from_until", "num_row", "base_slice", "base_columns", "nonbase_columns", ]:
        exec(f"{n} = property(lambda self: self._invariant.{n})", )

    def get_data_variant(
        self,
        vid: int = 0,
    ) -> _np.ndarray:
        """
        """
        return self._variants[vid].data

    get_data_array_variant = get_data_variant

    def remove_initial(self, /, ) -> None:
        """
        """
        remove = -self._invariant.min_max_shift[0]
        self.remove_periods_from_start(remove, )

    def remove_terminal(self, /, ) -> None:
        """
        """
        remove = self._invariant.min_max_shift[1]
        self.remove_periods_from_end(remove, )

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return self._invariant.create_name_to_row()

    def remove_periods_from_start(
        self,
        num_periods_to_remove: int,
        /,
    ) -> None:
        """
        """
        if num_periods_to_remove < 0:
            raise ValueError("Cannot remove a negative number of periods")
        self._invariant.remove_periods_from_start(num_periods_to_remove, )
        for v in self._variants:
            v.remove_periods_from_start(num_periods_to_remove, )

    def add_periods_to_end(
        self,
        num_periods_to_add: int,
        /,
    ) -> None:
        """
        """
        if num_periods_to_add < 0:
            raise ValueError("Cannot add a negative number of periodsk")
        self._invariant.add_periods_to_end(num_periods_to_add, )
        for v in self._variants:
            v.add_periods_to_end(num_periods_to_add, )

    def rescale_data(self, *args, **kwargs, ) -> None:
        """
        Rescale data in all variants by a common factor
        """
        for v in self._variants:
            v.rescale_data(*args, **kwargs, )

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove < 0:
            raise ValueError("Cannot remove negative number of columns")
        self._invariant.remove_periods_from_end(remove, )
        for v in self._variants:
            v.remove_periods_from_end(remove, )

    def logarithmize(self, /, ) -> None:
        """
        """
        logly_indexes = tuple(self.logly_indexes)
        if not logly_indexes:
            return
        for v in self._variants:
            v.logarithmize(logly_indexes, )

    def delogarithmize(self, /, ) -> None:
        """
        """
        logly_indexes = tuple(self.logly_indexes)
        if not logly_indexes:
            return
        for v in self._variants:
            v.delogarithmize(logly_indexes, )
    #]


def _get_extended_span(
    slatable: SlatableProtocol,
    base_span: Iterable[Period],
    /,
    prepend_initial: bool,
    append_terminal: bool,
) -> tuple[Iterable[Period], tuple[int, ...]]:
    """
    """
    base_span = tuple(t for t in base_span)
    num_base_periods = len(base_span)
    min_shift = slatable.max_lag if prepend_initial else 0
    max_shift = slatable.max_lead if append_terminal else 0
    min_base_date = min(base_span)
    max_base_date = max(base_span)
    start_date = min_base_date + min_shift
    end_date = max_base_date + max_shift
    base_columns = tuple(_dates.period_indexes(base_span, start_date, ))
    extended_dates = tuple(Span(start_date, end_date))
    return extended_dates, base_columns, min_shift, max_shift


def _slate_value_variant_iterator(
    value: Any,
    /,
    from_until: tuple[Period, Period],
) -> Iterator[Any]:
    """
    """
    #[
    if hasattr(value, "iter_data_variants_from_until"):
        return value.iter_data_variants_from_until(from_until, )
    elif isinstance(value, Iterable):
        return _iterators.exhaust_then_last(value, )
    else:
        return _iterators.exhaust_then_last([], value, )
    #]


def retrieve_vector_from_data_array(
    data: _np.ndarray,
    tokens: tuple[str, ...],
    column_zero: int,
    /,
) -> _np.ndarray:
    """
    """
    rows, columns = _incidences.rows_and_columns_from_tokens(tokens, column_zero, )
    return data[rows, columns]


def store_vector_in_data_array(
    vector: _np.ndarray,
    data: _np.ndarray,
    tokens: Iterable[_incidences.Incidence],
    column_zero: int,
) -> None:
    """
    """
    rows, columns = _incidences.rows_and_columns_from_tokens(tokens, column_zero, )
    data[rows, columns] = vector




"""
"""


#[
from __future__ import annotations

from ..conveniences import views as _views
from .. import dates as _dates
#]


class Inlay(_views.Mixin, ):
    """
    """
    #[

    def _get_first_line_view(self, /, ):
        """
        """
        shape = self.data.shape
        span_str = _dates.get_printable_span(self.start, self.end, )
        missing_str = "*" if self.has_missing else " "
        return f"Series {self.frequency.letter} {span_str}{missing_str}{shape[0]}×{shape[1]}"

    def _get_header_separator(self, /, ):
        """"
        """
        return ""

    def _get_content_view(self, /, ):
        """
        """
        return tuple(
            _get_series_row_str_(date, data_row, self._date_str_format, self._numeric_format, self._missing_str) 
            for row, (date, data_row) in enumerate(zip(self.span, self.data))
        )

    def _get_short_row_(self):
        """
        Create string representing rows skipped in the short view of times series
        """
        shape = self.data.shape
        date_str = ("{:"+self._date_str_format+"}").format(_views._VERTICAL_ELLIPSIS)
        value_format = "{:"+self._short_str_format+"}"
        data_str = "".join([value_format.format(_views._VERTICAL_ELLIPSIS)] * shape[1])
        return date_str + data_str

    #]


def _get_series_row_str_(date, data, date_str_format, numeric_format, missing_str: str):
    """
    String representing one row of time series data including the date
    """
    #[
    date_str = ("{:"+date_str_format+"}").format(date)
    value_format = "{:"+numeric_format+"}"
    data_str = "".join([value_format.format(v) for v in data])
    data_str = data_str.replace("nan", "{:>3}".format(missing_str))
    return date_str + data_str
    #]




"""
"""


#[
from __future__ import annotations

from typing import (Any, )
from collections.abc import (Iterable, )
import copy as _co

from .. import quantities as _quantities
from ..databoxes import main as _databoxes
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "parameters",
    )

    def __init__(
        self,
        parameter_names: Iterable[str],
        /,
    ) -> None:
        """
        """
        self.parameters = { n: None for n in parameter_names }

    def copy(self, /, ) -> Self:
        """
        """
        return _co.deepcopy(self, )

    def assign_from_dict_like(self, dict_like: dict[str, Any], ) -> None:
        """
        """
        existing_keys = set(self.parameters.keys())
        custom_keys = set(dict_like.keys())
        for n in (custom_keys & existing_keys):
            self.parameters[n] = dict_like[n]

    #]




"""
Incidence tokens and wrt tokens
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from numbers import (Real, )
from typing import (NamedTuple, Callable, Self, Protocol, TypeAlias, )
import operator as _op
import itertools as _it
import numpy as _np

from .. import quantities as _quantities
#]


_PRINT_TOKEN = "x[({qid},t{shift:+g})]"
_PRINT_ZERO_SHIFT_TOKEN = "x[({qid},t)]"
_PRINT_SHIFT = "[{shift:+g}]"
ZERO_SHIFT_TOKEN_PATTERN = r"x\[\((\d+),t\)\]"


class Token(NamedTuple, ):
    """
    Incidence of quantities in equations
    """
    #[

    qid: int
    shift: int

    def shifted(
        self: Self,
        by: int,
    ) -> Self:
        """
        Create a new Token with a shifted time
        """
        return Token(self.qid, self.shift+by, )

    def print(
        self,
        qid_to_name: dict[int, str],
    ) -> str:
        name = qid_to_name[self.qid]
        shift = _PRINT_SHIFT.format(shift=self.shift, ) if self.shift else ""
        return name + shift

    def print_xtring(self, /, ) -> str:
        return (
            _PRINT_TOKEN.format(qid=self.qid, shift=self.shift, )
            if self.shift else _PRINT_ZERO_SHIFT_TOKEN.format(qid=self.qid, )
        )

    #]


def get_max_shift(tokens: Iterable[Token]) -> int:
    return max(tok.shift for tok in tokens) if tokens else None


def get_min_shift(tokens: Iterable[Token]) -> int:
    return min(tok.shift for tok in tokens) if tokens else None


def get_max_qid(tokens: Iterable[Token]) -> int:
    return max(tok.qid for tok in tokens) if tokens else None


def generate_qids_from_tokens(tokens: Iterable[Token]) -> Iterable[int]:
    return (tok.qid for tok in tokens)


def get_some_shift_by_quantities(tokens: Iterable[Token], something: Callable, ) -> dict:
    """
    """
    key = _op.attrgetter("qid", )
    sorted_tokens = sorted(tokens, key=key, )
    return {
        qid: something(t.shift for t in tokens)
        for qid, tokens in _it.groupby(sorted_tokens, key=key, )
    }


def generate_tokens_of_kinds(tokens: Iterable[Token], qid_to_kind: dict, kinds: _quantities.QuantityKind) -> Iterable[Token]:
    return (tok for tok in tokens if qid_to_kind[tok.qid] in kinds)


def sort_tokens(tokens: Iterable[Token]) -> Iterable[Token]:
    """
    Sort tokens by shift and id
    """
    return sorted(tokens, key=lambda x: (-x.shift, x.qid))


def is_qid_in_tokens(tokens: Iterable[Token], qid: int) -> bool:
    return any(tok.qid == qid for tok in tokens)


def is_qid_zero_in_tokens(tokens: Iterable[Token], qid: int) -> bool:
    return any(tok.qid == qid and tok.shift == 0 for tok in tokens)


def generate_zero_shift_tokens_from_quantities(
    quantities: Iterable[_quantities.Quantity],
    *args, **kwargs,
) -> Iterable[Token]:
    """
    """
    quantities = _quantities.generate_quantities_of_kind(quantities, *args, **kwargs, )
    return ( Token(qid=q.id, shift=0, ) for q in quantities )


def print_tokens(
    tokens: Iterable[Token],
    /,
    qid_to_name: dict[int, str],
    qid_to_logly: dict[int, bool],
) -> Iterable[str]:
    """
    Create list of printed tokens
    """
    qid_to_logly = qid_to_logly or {}
    return [
        _quantities.wrap_logly(t.print(qid_to_name), qid_to_logly.get(t.qid, False), )
        for t in tokens
    ]


def rows_and_columns_from_tokens(
    tokens: Iterable[Token],
    column_zero: int,
) -> tuple[Iterable[int], Iterable[int]]:
    """
    """
    rows, columns = zip(*((t.qid, column_zero+t.shift) for t in tokens))
    return tuple(rows), tuple(columns)




"""
Implement SlatableProtocol
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )

from ..slatables import (Slatable, )

if TYPE_CHECKING:
    from typing import (Self, )
#]


_DEFAULT_RESIDUAL_VALUE = 0.0


class _Slatable(Slatable):
    """
    """
    #[

    @classmethod
    def for_simulate(
        klass,
        sequential,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass(**kwargs, )
        #
        self.max_lag = sequential.max_lag
        self.max_lead = sequential.max_lead
        #
        self.databox_names = sequential.all_names
        self.descriptions = None
        self.databox_validators = None
        #
        #
        self.fallbacks = {}
        self.overwrites = {}
        #
        #
        parameter_name_to_value = sequential.get_parameters(unpack_singleton=True, )
        if self.parameters_from_data:
            self.fallbacks.update(parameter_name_to_value, )
        else:
            self.overwrites.update(parameter_name_to_value, )
        #
        #
        residual_name_to_value = {
            name: _DEFAULT_RESIDUAL_VALUE
            for name in sequential.residual_names
        }
        if self.shocks_from_data:
            self.fallbacks.update(residual_name_to_value, )
        else:
            self.overwrites.update(residual_name_to_value, )
        #
        #
        self.qid_to_logly = {}
        #
        self.output_names = (
            sequential.lhs_names
            + sequential.rhs_only_names
            + sequential.residual_names
        )
        if self.parameters_from_data:
            self.output_names += sequential.parameter_names
        #
        return self


class Inlay:
    """
    """
    #[

    def get_slatable_for_simulation(self, **kwargs, ) -> _Slatable:
        """
        """
        return _Slatable.for_simulate(self, **kwargs, )





"""
Description mixin
"""


#[
from __future__ import annotations

import documark as _dm
#]


class DescriptionMixin:
    """
    """
    #[

    @_dm.reference(category="information", )
    def get_description(self, /, ) -> str:
        r"""
................................................................................


==Get the description attached an Iris Pie object==

    description = self.get_description()

### Input arguments ###

???+ input "self"
    An Iris Pie object from which to get the description.


### Returns ###

???+ returns "description"
    The description attached to the Iris Pie object.


................................................................................
        """
        return str(self.__description__ or "")

    @_dm.reference(category="information", )
    def set_description(self, description: str, /, ) -> None:
        r"""
................................................................................


==Set the description for an Iris Pie object==

    self.set_description(
        description,
    )


### Input arguments ###

???+ input "self"
    An Iris Pie object to which to attach the description.


???+ input "description"
    The description to attach to the Iris Pie object.


### Returns ###

This method modifies the Iris Pie object in place and returns `None`.


................................................................................
        """
        self.__description__ = str(description or "")

    #]




r"""
Multivariate linear least squares estimators
"""

#[

from __future__ import annotations

import numpy as _np

#]


__all__ = (
    "lstsq",
    "destandardize_lstsq",
    "standardize",
    "destandardize",
)


def lstsq(
    lhs: _np.ndarray,
    rhs: _np.ndarray,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    Mx = rhs @ rhs.T
    My = rhs @ lhs.T
    beta = _np.linalg.solve(Mx, My).T
    res = lhs - beta @ rhs
    return beta, res,


def destandardize_lstsq(bb, lhs, rhs, ) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    mean_lhs, std_lhs = lhs
    mean_rhs, std_rhs = rhs
    inv_sigma_lhs = _np.diag(1 / std_lhs)
    sigma_rhs = _np.diag(std_rhs)
    b = inv_sigma_lhs @ bb @ sigma_rhs
    c = mean_lhs - b @ mean_rhs
    return b, c,


def standardize(
    x: _np.ndarray,
) -> tuple[_np.ndarray, tuple[_np.ndarray, _np.ndarray]]:
    r"""
    """
    mean = _np.mean(x, axis=1, )
    std = _np.std(x, axis=1, mean=mean.reshape(-1, 1, ), )
    standardized = (x - mean.reshape(-1, 1, )) / std.reshape(-1, 1, )
    mean_std = (mean, std, )
    return standardized, mean_std


def destandardize(
    x: _np.ndarray,
    mean_std : tuple[_np.ndarray, _np.ndarray],
) -> _np.ndarray:
    r"""
    """
    mean, std = mean_std
    return x * std.reshape(-1, 1, ) + mean.reshape(-1, 1, ),




"""
"""


#[
from __future__ import absolute_import
#]


class Inlay:
    """
    """
    #[

    def get_parameter_names(self, /, ) -> tuple[str, ...]:
        r"""
        """
        return tuple(self.parameter_names)

    def get_residual_names(self, /, ) -> tuple[str, ...]:
        r"""
        """
        return tuple(self.residual_names)

    def get_equations(self, /, ) -> tuple[str, ...]:
        r"""
        """
        return tuple(e.human for e in self._invariant.equations)

    #]




r"""
Prior dummy observations
"""


#[

from __future__ import annotations

import numpy as _np
from numbers import Real

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import NoReturn

#]


__all__ = (
    "LittermanDummies",
)


class _VecAutoregDimsMixin:
    r"""
    """
    #[

    @property
    def num_endogenous(self, ) -> int:
        return self.dims[0]

    @property
    def num_exogenous(self, ) -> int:
        return self.dims[1]

    @property
    def has_intercept(self, ) -> bool:
        return bool(self.dims[2])

    @property
    def num_nonendogenous(self, ) -> int:
        return self.num_exogenous + int(self.has_intercept)

    @property
    def order(self, ) -> int:
        return self.dims[3]

    #]


class LittermanDummies(
    _VecAutoregDimsMixin,
):
    r"""
    """
    #[

    def __init__(
        self,
        dims: tuple[int, int, bool, int],
        rho: Real | _np.ndarray,
        mu: Real | _np.ndarray,
        kappa: Real,
    ) -> None:
        r"""
        """
        self.dims = dims
        self.rho = _ensure_array(rho, self.num_endogenous, )
        self.mu = _ensure_array(mu, self.num_endogenous, )
        self.kappa = kappa
        #
        self.num_dummies = self.num_endogenous * self.order
        self._populate_y0()
        self._populate_y1()
        self._populate_x()

    def _populate_y0(self, ) -> None:
        r"""
        """
        num_endogenous = self.num_endogenous
        order = self.order
        self.y0 = _np.hstack([
            _np.diag(self.mu * self.rho),
            _np.zeros((num_endogenous, num_endogenous * (order - 1)), ),
        ], dtype=float)

    def _populate_y1(self, ) -> None:
        r"""
        """
        order_multipliers = ( (i+1)**self.kappa for i in range(self.order) )
        self.y1 = _np.diag(_np.hstack([ self.mu * i for i in order_multipliers ]))

    def _populate_x(self, ) -> None:
        r"""
        """
        shape = (self.num_nonendogenous, self.num_dummies, )
        self.x = _np.zeros(shape, dtype=float, )

    #]


def _ensure_array(
    x: Real | _np.ndarray,
    num: int,
) -> _np.ndarray | NoReturn:
    r"""
    """
    #[
    shape = (num, )
    if isinstance(x, Real) or x.size == 1:
        x = _np.full(shape, x)
    if x.shape != shape:
        raise ValueError(f"Expected {shape} but got {x.shape}", )
    return x.astype(float, )
    #]





"""
"""

#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, )
from types import (EllipsisType, )
#]


class ItemMixin:
    """
    """
    #[

    def __getitem__(
        self,
        index: int | slice,
    ) -> Self:
        """
        """
        row_index, column_index = index[0], index[1]
        new = self.copy()
        if not _is_all(row_index):
            raise NotImplementedError
        if not _is_all(column_index):
            new._select_columns(new, column_index, )
        return new

    def _select_columns(
        self,
        column_index: int | slice | EllipsisType,
    ) -> Self:
        """
        """
        if isinstance(column_index, int):
            column_index = slice(column_index, column_index + 1, )
        self.base_span = self.base_span[column_index]
        for r in ("exogenized", "endogenized", "anticipated", ):
            register = getattr(self, f"_{r}_register")
            for k, v in register.items():
                register[k] = v[column_index]
    #]


def _is_all(index, /, ) -> bool:
    """
    """
    return index is ... or index == slice(None, )

